
from typing import ClassVar, Optional

from cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.BusinessObject import (
    ReadResponse,
    SchemaResponse,
)
from cherwell_pydantic_api.api import Connection
from cherwell_pydantic_api.bo.interfaces import ApiRequesterInterface
from cherwell_pydantic_api.bo.registry import BusinessObjectRegistry
from cherwell_pydantic_api.settings import InstanceSettingsBase, Settings
from cherwell_pydantic_api.types import BusObID, BusObRecID



class Instance(ApiRequesterInterface):
    _instances: ClassVar[dict[str, "Instance"]] = {}

    def __init__(self, instance_settings: InstanceSettingsBase):
        """Internal use only. Use Instance.use() instead."""
        self._settings = instance_settings.copy(
        )  # Take into account that the settings might change under our nose
        self.bo = BusinessObjectRegistry(self)
        self._connection = Connection(instance_settings)


    @property
    def settings(self) -> InstanceSettingsBase:
        return self._settings
    
    
    async def authenticate(self):
        return await self._connection.authenticate()


    async def get_busobid(self, busobname: str) -> Optional[BusObID]:
        return await self._connection.get_busobid(busobname)


    async def get_bo_schema(self, *, busobid: Optional[BusObID] = None, busobname: Optional[str] = None) -> Optional[SchemaResponse]:
        if busobid is None:
            if busobname is None:
                raise TypeError(
                    'get_bo_schema() missing 1 required argument: busobid or busobname')
            busobid = await self.get_busobid(busobname)
            if busobid is None:
                return None
        response = await self._connection.GetBusinessObjectSchemaV1(busobId=busobid, includerelationships=True)
        if not response.hasError:
            self.bo.register(response)
        return response


    async def get_bo(self, busobid: BusObID, *, busobrecid: Optional[BusObRecID] = None, publicid: Optional[str] = None) -> ReadResponse:
        if busobrecid is None:
            if publicid is None:
                raise TypeError(
                    'get_bo() missing 1 required argument: busobrecid or publicid')
            return await self._connection.GetBusinessObjectByPublicIdV1(busobid=busobid, publicid=publicid)
        return await self._connection.GetBusinessObjectByRecIdV1(busobid=busobid, busobrecid=busobrecid)


    @classmethod
    def use(cls, name: Optional[str] = None, *, instance_settings: Optional[InstanceSettingsBase] = None) -> "Instance":
        """Instantiate an Instance object for the given instance name, or return an existing Instance object if one already exists."""
        settings = Settings()
        if instance_settings is None:
            instance_settings = settings.get_instance(name)
        if instance_settings is None:
            raise ValueError(f"Instance {name} not found in settings")
        name = instance_settings.name
        if name not in cls._instances:
            cls._instances[name] = cls(instance_settings)
        return cls._instances[name]




from typing import Any, Dict, List, Literal, Optional, Union, AsyncIterable, Iterable
from pydantic import parse_obj_as
import cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.BusinessObject
import cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.Approval
from cherwell_pydantic_api.generated_api_utils import GeneratedInterfaceBase


class ApprovalInterface(GeneratedInterfaceBase):
    async def ActionApprovalV1(
        self,
        approvalRecId: str,
        approvalAction: str,
    ) -> cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.BusinessObject.SaveResponse:
        """Action an Approval

        Operation that actions an Approval Business Object. Use this method, passing either approve, abstain or deny to update the Business Object's state
         :param approvalRecId:
         :param approvalAction:
         :return: cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.BusinessObject.SaveResponse
        """
        self.validate_path_param(approvalRecId)
        self.validate_path_param(approvalAction)
        response = await self.post(f"/api/V1/approval/{approvalRecId}/{approvalAction}")
        return parse_obj_as(
            cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.BusinessObject.SaveResponse,
            response,
        )

    async def GetApprovalByRecIdV1(
        self,
        approvalRecId: str,
    ) -> cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.Approval.ApprovalReadResponse:
        """Get Approval

        Operation that returns an Approval Business Object.  Use the provided links to action the Approval
         :param approvalRecId:
         :return: cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.Approval.ApprovalReadResponse
        """
        self.validate_path_param(approvalRecId)
        response = await self.get(f"/api/V1/approval/{approvalRecId}")
        return parse_obj_as(
            cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.Approval.ApprovalReadResponse,
            response,
        )

    async def GetMyApprovalsV1(
        self,
    ) -> cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.Approval.GetApprovalsResponse:
        """Get all waiting Approvals for the current user

        Operation that returns a list of Approval Business Objects that are in a state of 'Waiting' for the current user.  Use the provided links to action the Approval
         :return: cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.Approval.GetApprovalsResponse
        """
        response = await self.get("/api/V1/getmyapprovals")
        return parse_obj_as(
            cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.Approval.GetApprovalsResponse,
            response,
        )

    async def GetMyPendingApprovalsV1(
        self,
    ) -> cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.Approval.GetApprovalsResponse:
        """Get all waiting approvals that were created by the current user

        Operation that returns a list of Approval Business Objects that are in a state of 'Waiting' that were created by the current user.  Use the provided links to action the Approval
         :return: cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.Approval.GetApprovalsResponse
        """
        response = await self.get("/api/V1/getmypendingapprovals")
        return parse_obj_as(
            cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.Approval.GetApprovalsResponse,
            response,
        )



#!/usr/bin/env python3

from typing import TYPE_CHECKING

import IPython
import traitlets



def ipython_shell():
    import cherwell_pydantic_api
    from cherwell_pydantic_api.settings import Settings
    from cherwell_pydantic_api.instance import Instance
    from cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts import (
        BusinessObject,
        Core,
        Searches,
    )

    _cfg = traitlets.config.Config() # type: ignore
    _cfg.InteractiveShellApp.exec_lines = [
        "cw_inst = Instance.use()",
        "await cw_inst.authenticate()",
        "await cw_inst.get_bo_schema(busobname='Ticket')",
    ]
    IPython.start_ipython(config=_cfg, user_ns=locals())
    
    if TYPE_CHECKING:
        cherwell_pydantic_api # type: ignore
        Core # type: ignore
        BusinessObject # type: ignore
        Searches # type: ignore
        Settings # type: ignore
        Instance # type: ignore


if __name__ == '__main__':
    ipython_shell()



# generated by datamodel-codegen:
#   filename:  swagger.json
#   timestamp: 2023-03-19T19:24:29+00:00

from __future__ import annotations

from typing import Optional

import cherwell_pydantic_api.types as ct
from cherwell_pydantic_api.generated_api_utils import ApiBaseModel

from .. import CherwellObjectID


class CherwellNameValuePair(ApiBaseModel):
    name: Optional[str] = None
    valueObject: Optional[CherwellObjectID] = None
    valueString: Optional[str] = None
    category: Optional[str] = None
    description: Optional[str] = None
    displayShowsValue: Optional[bool] = None
    specialUseFlag: Optional[bool] = None
    displayString: Optional[str] = None



from typing import Any, Dict, List, Literal, Optional, Union, AsyncIterable, Iterable
from pydantic import parse_obj_as
import cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.Core
import cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.Searches
from cherwell_pydantic_api.generated_api_utils import GeneratedInterfaceBase


class SearchesInterface(GeneratedInterfaceBase):
    async def GetQuickSearchConfigurationForBusObsV1(
        self,
        dataRequest: cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.Searches.QuickSearchConfigurationRequest,
    ) -> cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.Searches.QuickSearchConfigurationResponse:
        """Get a Quick Search from a list of Business Object IDs

        Operation to build a Quick Search configuration that you can use to execute a Quick Search for multiple Business Objects. The configuration  includes supplied Business Object IDs and specific search items with the following options. Use the Option Key to determine if you can change the options.


        ChangedOption

        NonFinalStateOption

        SearchAnyWordsOption

        SearchAttachmentsOption

        SearchRelatedOption

        SortByOption


        Option Key:

        0 = None (Not selected and cannot select.)

        1 = Use (Selected and cannot clear.)

        2 = Display (Not selected and can select.)

        3 = UseAndDisplay (Selected and can clear.)


        SearchTargetType:

        0 = BusOb (Business Object)

        1 = DocRepository (Document Repository)
         :param dataRequest: Request containing the Business Object IDs list.
         :return: cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.Searches.QuickSearchConfigurationResponse
        """
        response = await self.post_body(
            "/api/V1/getquicksearchconfigurationforbusobs",
            content=dataRequest.json(exclude_unset=True, by_alias=True),
        )
        return parse_obj_as(
            cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.Searches.QuickSearchConfigurationResponse,
            response,
        )

    async def GetQuickSearchConfigurationForBusObsWithViewRightsV1(
        self,
    ) -> cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.Searches.QuickSearchConfigurationResponse:
        """Get a Quick Search by Business Objects with view rights

        Operation to get a Quick Search configuration that you can use to execute a Quick Search based the current user's Business Object view rights. The configuration  includes supplied Business Object IDs and specific search items with the following options. Use the Option Key to determine if you can change the options.

        ChangedOption

        NonFinalStateOption

        SearchAnyWordsOption

        SearchAttachmentsOption

        SearchRelatedOption

        SortByOption


        Option Key:

        0 = None (Not selected and cannot select.)

        1 = Use (Selected and cannot clear.)

        2 = Display (Not selected and can select.)

        3 = UseAndDisplay (Selected and can clear.)


        SearchTargetType:

        0 = BusOb (Business Object)

        1 = DocRepository (Document Repository)
         :return: cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.Searches.QuickSearchConfigurationResponse
        """
        response = await self.get(
            "/api/V1/getquicksearchconfigurationforbusobswithviewrights"
        )
        return parse_obj_as(
            cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.Searches.QuickSearchConfigurationResponse,
            response,
        )

    async def GetQuickSearchResultsV1(
        self,
        dataRequest: cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.Searches.QuickSearchRequest,
        includeLinks: Optional[bool] = None,
    ) -> cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.Searches.SimpleResultsList:
        """Execute a Quick Search from a list of Business Object IDs and search text

        Operation to execute a Quick Search using a list of Business Object IDs and search text.
         :param dataRequest: Request object listing Business Object IDs and search text. Leave out the entire Business Object IDs parameter and all configured quick search Business Objects will be searched.
         :param includeLinks: Flag to include hyperlinks in results. Default is false.
         :return: cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.Searches.SimpleResultsList
        """
        params = {}
        if includeLinks is not None:
            params["includeLinks"] = includeLinks
        response = await self.post_body(
            "/api/V1/getquicksearchresults",
            params=params,
            content=dataRequest.json(exclude_unset=True, by_alias=True),
        )
        return parse_obj_as(
            cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.Searches.SimpleResultsList,
            response,
        )

    async def GetQuickSearchSpecificResultsV1(
        self,
        dataRequest: cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.Searches.QuickSearchSpecificRequest,
        includeSchema: Optional[bool] = None,
        includeLocationFields: Optional[bool] = None,
        includeLinks: Optional[bool] = None,
    ) -> cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.Searches.SearchResultsTableResponse:
        """Execute a Quick Search on a specific Business Object

        Operation to execute a Quick Search for a specific Business Object ID. Use "Get a Quick Search from a list of Business Object IDs" to find values for specific search item options, such as NonFinalStateOption.
         :param dataRequest: Request object containing the parameters for specific Business Object Quick Search execution.
         :param includeSchema: Flag to include the schema for the results.
         :param includeLocationFields: Flag to include location fields in the results.
         :param includeLinks: Flag to include hyperlinks in results. Default is false.
         :return: cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.Searches.SearchResultsTableResponse
        """
        params = {}
        if includeSchema is not None:
            params["includeSchema"] = includeSchema
        if includeLocationFields is not None:
            params["includeLocationFields"] = includeLocationFields
        if includeLinks is not None:
            params["includeLinks"] = includeLinks
        response = await self.post_body(
            "/api/V1/getquicksearchspecificresults",
            params=params,
            content=dataRequest.json(exclude_unset=True, by_alias=True),
        )
        return parse_obj_as(
            cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.Searches.SearchResultsTableResponse,
            response,
        )

    async def GetQuickSearchSpecificResultsV2(
        self,
        dataRequest: cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.Searches.QuickSearchSpecificRequest,
        includeSchema: Optional[bool] = None,
        includeLocationFields: Optional[bool] = None,
        includeLinks: Optional[bool] = None,
    ) -> cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.Searches.QuickSearchResponse:
        """Execute a Quick Search on a specific Business Object

        Operation to execute a Quick Search for a specific Business Object ID. Use "Get a Quick Search from a list of Business Object IDs" to find values for specific search item options, such as NonFinalStateOption.
         :param dataRequest: Request object containing the parameters for specific Business Object Quick Search execution.
         :param includeSchema: Flag to include the schema for the results.
         :param includeLocationFields: Flag to include location fields in the results.
         :param includeLinks: Flag to include hyperlinks in results. Default is false.
         :return: cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.Searches.QuickSearchResponse
        """
        params = {}
        if includeSchema is not None:
            params["includeSchema"] = includeSchema
        if includeLocationFields is not None:
            params["includeLocationFields"] = includeLocationFields
        if includeLinks is not None:
            params["includeLinks"] = includeLinks
        response = await self.post_body(
            "/api/V2/getquicksearchspecificresults",
            params=params,
            content=dataRequest.json(exclude_unset=True, by_alias=True),
        )
        return parse_obj_as(
            cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.Searches.QuickSearchResponse,
            response,
        )

    async def GetSearchItemsByAssociation_Scope_ScopeOwner_FolderV1(
        self,
        association: str,
        scope: str,
        scopeowner: str,
        folder: str,
        links: Optional[bool] = None,
    ) -> cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.Searches.SearchItemResponse:
        """Get all saved searches by Folder ID

        Operation that returns a tree of saved queries, including scope, search name, IDs, and location within the tree.
         :param association: Use to filter results by Business Object association ID.
         :param scope: Use to filter results by scope name or ID.
         :param scopeowner: Use to filter results by scope owner ID.
         :param folder: Use to filter results by Search Group folder ID.
         :param links: Flag to include hyperlinks in results. Default is false.
         :return: cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.Searches.SearchItemResponse
        """
        self.validate_path_param(association)
        self.validate_path_param(scope)
        self.validate_path_param(scopeowner)
        self.validate_path_param(folder)
        params = {}
        if links is not None:
            params["links"] = links
        response = await self.get(
            f"/api/V1/getsearchitems/association/{association}/scope/{scope}/scopeowner/{scopeowner}/folder/{folder}",
            params=params,
        )
        return parse_obj_as(
            cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.Searches.SearchItemResponse,
            response,
        )

    async def GetSearchItemsByAssociation_Scope_ScopeOwner_FolderV2(
        self,
        association: str,
        scope: str,
        scopeowner: str,
        folder: str,
        links: Optional[bool] = None,
    ) -> cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.Core.ManagerData:
        """Get all saved searches by Folder ID

        Operation that returns a tree of saved queries, including scope, search name, IDs, and location within the tree.
         :param association: Use to filter results by Business Object association ID.
         :param scope: Use to filter results by scope name or ID.
         :param scopeowner: Use to filter results by scope owner ID.
         :param folder: Use to filter results by Search Group folder ID.
         :param links: Flag to include hyperlinks in results. Default is false.
         :return: cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.Core.ManagerData
        """
        self.validate_path_param(association)
        self.validate_path_param(scope)
        self.validate_path_param(scopeowner)
        self.validate_path_param(folder)
        params = {}
        if links is not None:
            params["links"] = links
        response = await self.get(
            f"/api/V2/getsearchitems/association/{association}/scope/{scope}/scopeowner/{scopeowner}/folder/{folder}",
            params=params,
        )
        return parse_obj_as(
            cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.Core.ManagerData,
            response,
        )

    async def GetSearchItemsByAssociation_Scope_ScopeOwnerV1(
        self,
        association: str,
        scope: str,
        scopeowner: str,
        links: Optional[bool] = None,
    ) -> cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.Searches.SearchItemResponse:
        """Get all saved searches by scope owner (sub scope)

        Operation that returns a tree of saved queries, including scope, search name, IDs, and location within the tree.
         :param association: Use to filter results by Business Object association ID.
         :param scope: Use to filter results by scope name or ID.
         :param scopeowner: Use to filter results by scope owner ID.
         :param links: Flag to include hyperlinks in results. Default is false.
         :return: cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.Searches.SearchItemResponse
        """
        self.validate_path_param(association)
        self.validate_path_param(scope)
        self.validate_path_param(scopeowner)
        params = {}
        if links is not None:
            params["links"] = links
        response = await self.get(
            f"/api/V1/getsearchitems/association/{association}/scope/{scope}/scopeowner/{scopeowner}",
            params=params,
        )
        return parse_obj_as(
            cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.Searches.SearchItemResponse,
            response,
        )

    async def GetSearchItemsByAssociation_Scope_ScopeOwnerV2(
        self,
        association: str,
        scope: str,
        scopeowner: str,
        links: Optional[bool] = None,
    ) -> cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.Core.ManagerData:
        """Get all saved searches by scope owner (sub scope)

        Operation that returns a tree of saved queries, including scope, search name, IDs, and location within the tree.
         :param association: Use to filter results by Business Object association ID.
         :param scope: Use to filter results by scope name or ID.
         :param scopeowner: Use to filter results by scope owner ID.
         :param links: Flag to include hyperlinks in results. Default is false.
         :return: cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.Core.ManagerData
        """
        self.validate_path_param(association)
        self.validate_path_param(scope)
        self.validate_path_param(scopeowner)
        params = {}
        if links is not None:
            params["links"] = links
        response = await self.get(
            f"/api/V2/getsearchitems/association/{association}/scope/{scope}/scopeowner/{scopeowner}",
            params=params,
        )
        return parse_obj_as(
            cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.Core.ManagerData,
            response,
        )

    async def GetSearchItemsByAssociation_ScopeV1(
        self,
        association: str,
        scope: str,
        links: Optional[bool] = None,
    ) -> cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.Searches.SearchItemResponse:
        """Get all saved searches by scope

        Operation that returns a tree of saved queries, including scope, search name, IDs, and location within the tree.
         :param association: Use to filter results by Business Object association ID.
         :param scope: Use to filter results by scope name or ID.
         :param links: Flag to include hyperlinks in results. Default is false.
         :return: cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.Searches.SearchItemResponse
        """
        self.validate_path_param(association)
        self.validate_path_param(scope)
        params = {}
        if links is not None:
            params["links"] = links
        response = await self.get(
            f"/api/V1/getsearchitems/association/{association}/scope/{scope}",
            params=params,
        )
        return parse_obj_as(
            cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.Searches.SearchItemResponse,
            response,
        )

    async def GetSearchItemsByAssociation_ScopeV2(
        self,
        association: str,
        scope: str,
        links: Optional[bool] = None,
    ) -> cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.Core.ManagerData:
        """Get all saved searches by scope

        Operation that returns a tree of saved queries, including scope, search name, IDs, and location within the tree.
         :param association: Use to filter results by Business Object association ID.
         :param scope: Use to filter results by scope name or ID.
         :param links: Flag to include hyperlinks in results. Default is false.
         :return: cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.Core.ManagerData
        """
        self.validate_path_param(association)
        self.validate_path_param(scope)
        params = {}
        if links is not None:
            params["links"] = links
        response = await self.get(
            f"/api/V2/getsearchitems/association/{association}/scope/{scope}",
            params=params,
        )
        return parse_obj_as(
            cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.Core.ManagerData,
            response,
        )

    async def GetSearchItemsByAssociationV1(
        self,
        association: str,
        links: Optional[bool] = None,
    ) -> cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.Searches.SearchItemResponse:
        """Get all saved searches by Business Object association

        Operation that returns a tree of saved queries, including scope, search name, IDs, and location within the tree.
         :param association: Use to filter results by Business Object association ID.
         :param links: Flag to include hyperlinks in results. Default is false.
         :return: cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.Searches.SearchItemResponse
        """
        self.validate_path_param(association)
        params = {}
        if links is not None:
            params["links"] = links
        response = await self.get(
            f"/api/V1/getsearchitems/association/{association}", params=params
        )
        return parse_obj_as(
            cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.Searches.SearchItemResponse,
            response,
        )

    async def GetSearchItemsByAssociationV2(
        self,
        association: str,
        links: Optional[bool] = None,
    ) -> cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.Core.ManagerData:
        """Get all saved searches by Business Object association

        Operation that returns a tree of saved queries, including scope, search name, IDs, and location within the tree.
         :param association: Use to filter results by Business Object association ID.
         :param links: Flag to include hyperlinks in results. Default is false.
         :return: cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.Core.ManagerData
        """
        self.validate_path_param(association)
        params = {}
        if links is not None:
            params["links"] = links
        response = await self.get(
            f"/api/V2/getsearchitems/association/{association}", params=params
        )
        return parse_obj_as(
            cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.Core.ManagerData,
            response,
        )

    async def GetSearchItemsV1(
        self,
        links: Optional[bool] = None,
    ) -> cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.Searches.SearchItemResponse:
        """Get all saved searches by default Business Object association

        Operation that returns a tree of saved queries, including scope, search name, IDs, and location within the tree.
         :param links: Flag to include hyperlinks in results. Default is false.
         :return: cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.Searches.SearchItemResponse
        """
        params = {}
        if links is not None:
            params["links"] = links
        response = await self.get("/api/V1/getsearchitems", params=params)
        return parse_obj_as(
            cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.Searches.SearchItemResponse,
            response,
        )

    async def GetSearchItemsV2(
        self,
        links: Optional[bool] = None,
    ) -> cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.Core.ManagerData:
        """Get all saved searches by default Business Object association

        Operation that returns a tree of saved queries, including scope, search name, IDs, and location within the tree.
         :param links: Flag to include hyperlinks in results. Default is false.
         :return: cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.Core.ManagerData
        """
        params = {}
        if links is not None:
            params["links"] = links
        response = await self.get("/api/V2/getsearchitems", params=params)
        return parse_obj_as(
            cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.Core.ManagerData,
            response,
        )

    async def GetSearchResultsAdHocV1(
        self,
        dataRequest: cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.Searches.SearchResultsRequest,
    ) -> cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.Searches.SearchResultsResponse:
        """Run an ad-hoc search

        Operation that runs an ad-hoc Business Object search. To execute a search with Prompts, the PromptId and Value are required in the Prompt request object.

        PromptType is a FieldSubType enum as described below:
        FieldSubType
        None = 0
        Text = 1
        Number = 2
        DateTime = 3
        Logical = 4
        Binary = 5
        DateOnly = 6
        TimeOnly = 7
        Json = 8
        JsonArray = 9
        Xml = 10
        XmlCollection = 11
        TimeValue = 12

         :param dataRequest: Request object to specify search parameters.
         :return: cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.Searches.SearchResultsResponse
        """
        response = await self.post_body(
            "/api/V1/getsearchresults",
            content=dataRequest.json(exclude_unset=True, by_alias=True),
        )
        return parse_obj_as(
            cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.Searches.SearchResultsResponse,
            response,
        )

    async def GetSearchResultsByIdV1(
        self,
        association: str,
        scope: str,
        scopeowner: str,
        searchid: str,
        searchTerm: Optional[str] = None,
        pagenumber: Optional[int] = None,
        pagesize: Optional[int] = None,
        includeschema: Optional[bool] = None,
        resultsAsSimpleResultsList: Optional[bool] = None,
    ) -> cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.Searches.SearchResultsResponse:
        """Run a saved search by internal ID

        Operation that returns the paged results of a saved search. When the search contains Prompts, the response contains the Prompt. Send the Prompt and the original operation parameters to  SearchResultsRequest to the getsearchresults ad-hoc http post operation.

        PromptType is a FieldSubType enum as described below:
        FieldSubType
        None = 0
        Text = 1
        Number = 2
        DateTime = 3
        Logical = 4
        Binary = 5
        DateOnly = 6
        TimeOnly = 7
        Json = 8
        JsonArray = 9
        Xml = 10
        XmlCollection = 11
        TimeValue = 12

         :param association: Specify the Business Object association ID for the saved search.
         :param scope: Specify the scope name or ID for the saved search.
         :param scopeowner: Specify the scope owner ID for the saved search. Use (None) when no scope owner exists.
         :param searchid: Specify the internal ID for the saved search. Use "Run a saved search by name" if you do not have the internal ID.
         :param searchTerm: Specify search text filter the results. Example: Use "Service Request" to filter Incident results to include only service requests.
         :param pagenumber: Specify the page number of the result set to return.
         :param pagesize: Specify the number of rows to return per page.
         :param includeschema: Use to include the table schema of the saved search. If false, results contain the fieldid and field value without field information. Default is false.
         :param resultsAsSimpleResultsList: Indicates if the results should be returned in a simple results list format or a table format. Default is a table format.
         :return: cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.Searches.SearchResultsResponse
        """
        self.validate_path_param(association)
        self.validate_path_param(scope)
        self.validate_path_param(scopeowner)
        self.validate_path_param(searchid)
        params = {}
        if searchTerm is not None:
            params["searchTerm"] = searchTerm
        if pagenumber is not None:
            params["pagenumber"] = pagenumber
        if pagesize is not None:
            params["pagesize"] = pagesize
        if includeschema is not None:
            params["includeschema"] = includeschema
        if resultsAsSimpleResultsList is not None:
            params["resultsAsSimpleResultsList"] = resultsAsSimpleResultsList
        response = await self.get(
            f"/api/V1/getsearchresults/association/{association}/scope/{scope}/scopeowner/{scopeowner}/searchid/{searchid}",
            params=params,
        )
        return parse_obj_as(
            cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.Searches.SearchResultsResponse,
            response,
        )

    async def GetSearchResultsByNameV1(
        self,
        association: str,
        scope: str,
        scopeowner: str,
        searchname: str,
        searchTerm: Optional[str] = None,
        pagenumber: Optional[int] = None,
        pagesize: Optional[int] = None,
        includeschema: Optional[bool] = None,
        resultsAsSimpleResultsList: Optional[bool] = None,
    ) -> cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.Searches.SearchResultsResponse:
        """Run a saved search by name

        Operation that returns the paged results of a saved search. When the search contains Prompts, the response contains the Prompt. Send the Prompt and the original operation parameters to  SearchResultsRequest to the getsearchresults ad-hoc http post operation.

        PromptType is a FieldSubType enum as described below:
        FieldSubType
        None = 0
        Text = 1
        Number = 2
        DateTime = 3
        Logical = 4
        Binary = 5
        DateOnly = 6
        TimeOnly = 7
        Json = 8
        JsonArray = 9
        Xml = 10
        XmlCollection = 11
        TimeValue = 12

         :param association: Specify the Business Object association ID for the saved search.
         :param scope: Specify the scope name or ID for the saved search.
         :param scopeowner: Specify the scope owner ID for the saved search. Use (None) when no scope owner exists.
         :param searchname: Specify the name of the saved search.
         :param searchTerm: Specify search text filter the results. Example: Use "Service Request" to filter Incident results to include only service requests.
         :param pagenumber: Specify the page number of the result set to return.
         :param pagesize: Specify the number of rows to return per page.
         :param includeschema: Use to include the table schema of the saved search. If false, results contain the fieldid and field value without field information. Default is false.
         :param resultsAsSimpleResultsList: Indicates if the results should be returned in a simple results list format or a table format. Default is a table format.
         :return: cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.Searches.SearchResultsResponse
        """
        self.validate_path_param(association)
        self.validate_path_param(scope)
        self.validate_path_param(scopeowner)
        self.validate_path_param(searchname)
        params = {}
        if searchTerm is not None:
            params["searchTerm"] = searchTerm
        if pagenumber is not None:
            params["pagenumber"] = pagenumber
        if pagesize is not None:
            params["pagesize"] = pagesize
        if includeschema is not None:
            params["includeschema"] = includeschema
        if resultsAsSimpleResultsList is not None:
            params["resultsAsSimpleResultsList"] = resultsAsSimpleResultsList
        response = await self.get(
            f"/api/V1/getsearchresults/association/{association}/scope/{scope}/scopeowner/{scopeowner}/searchname/{searchname}",
            params=params,
        )
        return parse_obj_as(
            cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.Searches.SearchResultsResponse,
            response,
        )

    async def GetSearchResultsExportAdHocV1(
        self,
        dataRequest: cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.Searches.ExportSearchResultsRequest,
    ) -> str:
        """Export an ad-hoc search

        Operation that returns an ad-hoc search in a specified export format: 0=CSV, 1=Excel, 2=Tab, 3=Word, 4=Custom Separator, 5=Simple JSON. To execute a search with Prompts, the PromptId and Value are required in the Prompt request object.

        PromptType is a FieldSubType enum as described below:
        FieldSubType
        None = 0
        Text = 1
        Number = 2
        DateTime = 3
        Logical = 4
        Binary = 5
        DateOnly = 6
        TimeOnly = 7
        Json = 8
        JsonArray = 9
        Xml = 10
        XmlCollection = 11
        TimeValue = 12

         :param dataRequest: Request object to specify search parameters and export format.
         :return: str"""
        return await self.post_body(
            "/api/V1/getsearchresultsexport",
            content=dataRequest.json(exclude_unset=True, by_alias=True),
        )

    async def GetSearchResultsExportByIdV1(
        self,
        association: str,
        scope: str,
        scopeowner: str,
        searchid: str,
        exportformat: Literal["CSV", "Excel", "Tab", "Word", "CustomSeparator", "Json"],
        searchTerm: Optional[str] = None,
        pagenumber: Optional[int] = None,
        pagesize: Optional[int] = None,
    ) -> str:
        """Export a saved search by ID

        Operation that returns the paged results of a saved search in a specified format. When the search contains Prompts, the response contains the Prompt. Send the Prompt and the original operation parameters to  SearchResultsRequest to the getsearchresultsexport ad-hoc http post operation.

        PromptType is a FieldSubType enum as described below:
        FieldSubType
        None = 0
        Text = 1
        Number = 2
        DateTime = 3
        Logical = 4
        Binary = 5
        DateOnly = 6
        TimeOnly = 7
        Json = 8
        JsonArray = 9
        Xml = 10
        XmlCollection = 11
        TimeValue = 12

         :param association: Specify the Business Object association ID for the saved search.
         :param scope: Specify the scope name or ID for the saved search.
         :param scopeowner: Specify the scope owner ID for the saved search. Use (None) when no scope owner exists.
         :param searchid: Specify the internal ID for the saved search. Use "Run a saved search by name" if you do not have the internal ID.
         :param exportformat: Specify the format of the export
         :param searchTerm: Specify search text filter the results. Example: Use "Service Request" to filter Incident results to include only service requests.
         :param pagenumber: Specify the page number of the result set to return.
         :param pagesize: Specify the number of rows to return per page.
         :return: str"""
        self.validate_path_param(association)
        self.validate_path_param(scope)
        self.validate_path_param(scopeowner)
        self.validate_path_param(searchid)
        self.validate_path_param(exportformat)
        params = {}
        if searchTerm is not None:
            params["searchTerm"] = searchTerm
        if pagenumber is not None:
            params["pagenumber"] = pagenumber
        if pagesize is not None:
            params["pagesize"] = pagesize
        return await self.get(
            f"/api/V1/getsearchresultsexport/association/{association}/scope/{scope}/scopeowner/{scopeowner}/searchid/{searchid}/exportformat/{exportformat}",
            params=params,
        )

    async def GetSearchResultsExportByNameV1(
        self,
        association: str,
        scope: str,
        scopeowner: str,
        searchname: str,
        exportformat: Literal["CSV", "Excel", "Tab", "Word", "CustomSeparator", "Json"],
        searchTerm: Optional[str] = None,
        pagenumber: Optional[int] = None,
        pagesize: Optional[int] = None,
    ) -> str:
        """Export a saved search by name

        Operation that returns the paged results of a saved search in a specified format. When the search contains Prompts, the response contains the Prompt. Send the Prompt and the original operation parameters to  SearchResultsRequest to the getsearchresultsexport ad-hoc http post operation.

        PromptType is a FieldSubType enum as described below:
        FieldSubType
        None = 0
        Text = 1
        Number = 2
        DateTime = 3
        Logical = 4
        Binary = 5
        DateOnly = 6
        TimeOnly = 7
        Json = 8
        JsonArray = 9
        Xml = 10
        XmlCollection = 11
        TimeValue = 12

         :param association: Specify the Business Object association ID for the saved search.
         :param scope: Specify the scope name or ID for the saved search.
         :param scopeowner: Specify the scope owner ID for the saved search. Use (None) when no scope owner exists.
         :param searchname: Specify the name of the saved search.
         :param exportformat: Specify the format of the export
         :param searchTerm: Specify search text filter the results. Example: Use "Service Request" to filter Incident results to include only service requests.
         :param pagenumber: Specify the page number of the result set to return.
         :param pagesize: Specify the number of rows to return per page.
         :return: str"""
        self.validate_path_param(association)
        self.validate_path_param(scope)
        self.validate_path_param(scopeowner)
        self.validate_path_param(searchname)
        self.validate_path_param(exportformat)
        params = {}
        if searchTerm is not None:
            params["searchTerm"] = searchTerm
        if pagenumber is not None:
            params["pagenumber"] = pagenumber
        if pagesize is not None:
            params["pagesize"] = pagesize
        return await self.get(
            f"/api/V1/getsearchresultsexport/association/{association}/scope/{scope}/scopeowner/{scopeowner}/searchname/{searchname}/exportformat/{exportformat}",
            params=params,
        )

    async def GetSearchResultsAsStringByNameV1(
        self,
        scope: str,
        associationName: str,
        searchName: str,
        scopeOwner: Optional[str] = None,
    ) -> List[Dict]:
        """Get results of a saved search

        Operation that returns the results of a saved search in JSON format.

                        This API is protected by a rate limiter and will reject any requests sent from an IP Address when a certain threshold of active concurrent requests has been hit.

                        This value can be configured by the Max Concurrent Requests configuration value in the Web API config.

                        Once this limit has been reached, all subsequent requests will receive a status code of 429 (Too Many Requests).

         :param scope: Specify the scope name for the saved search.
         :param associationName: Specify the Business Object association Name for the saved search.
         :param searchName: Specify the name of the saved search.
         :param scopeOwner: Specify the scope owner ID for the saved search. Use (None) when no scope owner exists.
         :return: List[Dict]"""
        self.validate_path_param(scope)
        self.validate_path_param(associationName)
        self.validate_path_param(searchName)
        params = {}
        if scopeOwner is not None:
            params["scopeOwner"] = scopeOwner
        return await self.get(
            f"/api/V1/storedsearches/{scope}/{associationName}/{searchName}",
            params=params,
        )

    async def GetSearchResultsAsStringByNameV2(
        self,
        scope: str,
        associationName: str,
        searchName: str,
        scopeOwner: Optional[str] = None,
    ) -> List[Dict]:
        """Get results of a saved search

        Operation that returns the results of a saved search in JSON format.

        This API is protected by a rate limiter and will reject any requests sent from an IP Address when a certain threshold of active concurrent requests has been hit.

        This value can be configured by the Max Concurrent Requests configuration value in the Web API config.

        Once this limit has been reached, all subsequent requests will receive a status code of 429 (Too Many Requests).

        This version is not subject to row limits and will return the entire result set of the stored search.

         :param scope: Specify the scope name for the saved search.
         :param associationName: Specify the Business Object association Name for the saved search.
         :param searchName: Specify the name of the saved search.
         :param scopeOwner: Specify the scope owner ID for the saved search. Use (None) when no scope owner exists.
         :return: List[Dict]"""
        self.validate_path_param(scope)
        self.validate_path_param(associationName)
        self.validate_path_param(searchName)
        params = {}
        if scopeOwner is not None:
            params["scopeOwner"] = scopeOwner
        return await self.get(
            f"/api/V2/storedsearches/{scope}/{associationName}/{searchName}",
            params=params,
        )

    async def GetSearchResultsAsStringByIdV2(
        self,
        searchRequest: cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.Searches.StoredSearchRequest,
    ) -> cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.Searches.StoredSearchResults:
        """Get results of a saved search

        Operation that returns the results of a saved search in JSON format.

        This API is protected by a rate limiter and will reject any requests sent from an IP Address when a certain threshold of active concurrent requests has been hit.

        This value can be configured by the Max Concurrent Requests configuration value in the Web API config.

        Once this limit has been reached, all subsequent requests will receive a status code of 429 (Too Many Requests).

        This version is not subject to row limits and will return the entire result set of the stored search.

         :param searchRequest: Request object to specify search parameters.
         :return: cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.Searches.StoredSearchResults
        """
        response = await self.post_body(
            "/api/V2/storedsearches",
            content=searchRequest.json(exclude_unset=True, by_alias=True),
        )
        return parse_obj_as(
            cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.Searches.StoredSearchResults,
            response,
        )



from pathlib import Path
from cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.BusinessObject import ReadResponse
from cherwell_pydantic_api.bo.interfaces import ApiRequesterInterface
from cherwell_pydantic_api.bo.pydantic_model_generator import PydanticModelGenerator
from cherwell_pydantic_api.bo.valid_schema import ValidSchema



class BusinessObjectWrapper:
    """Wrap the SchemaResponse object to provide a more convenient interface to the business object schema.
    Terminology note: a "Business Object" is a Cherwell term for what I would call a "class" or "type". So each instance
    of BusinessObjectWrapper represents a type of business object."""

    def __init__(self, schema: ValidSchema, instance: ApiRequesterInterface):
        self._schema = schema
        self._instance = instance

    async def get(self, publicid: str) -> ReadResponse:
        return await self._instance.get_bo(self._schema.busObId, publicid=publicid)

    def generate_model_file(self) -> Path:
        generator = PydanticModelGenerator(self._instance.settings)
        return generator.generate_model_file(self._schema)
    


# generated by datamodel-codegen:
#   filename:  swagger.json
#   timestamp: 2023-03-19T19:24:29+00:00

from __future__ import annotations

from enum import Enum
from typing import List, Optional

import cherwell_pydantic_api.types as ct
from cherwell_pydantic_api.generated_api_utils import ApiBaseModel, HttpStatusCode


class GetStagesResponseStages(ApiBaseModel):
    id: Optional[str] = None
    isFinal: Optional[bool] = None
    name: Optional[str] = None


import cherwell_pydantic_api.types as ct


class GetStatusesResponseStatuses(ApiBaseModel):
    id: Optional[str] = None
    isInitial: Optional[bool] = None
    name: Optional[str] = None
    stageId: Optional[str] = None


import cherwell_pydantic_api.types as ct


class GetTransitionsResponseTransition(ApiBaseModel):
    id: Optional[str] = None
    name: Optional[str] = None
    fromStatusId: Optional[str] = None
    toStatusId: Optional[str] = None


import cherwell_pydantic_api.types as ct


class GetRecordStatusResponse(ApiBaseModel):
    id: Optional[str] = None
    name: Optional[str] = None
    errorCode: Optional[str] = None
    errorMessage: Optional[str] = None
    hasError: Optional[bool] = None
    httpStatusCode: Optional[HttpStatusCode] = None


import cherwell_pydantic_api.types as ct


class GetTransitionOptionsResponseTransition(ApiBaseModel):
    name: Optional[str] = None
    id: Optional[str] = None
    isAvailable: Optional[bool] = None
    criteria: Optional[List[str]] = None


import cherwell_pydantic_api.types as ct


class TransitionRecordRequest(ApiBaseModel):
    transitionOptionId: Optional[str] = None


import cherwell_pydantic_api.types as ct


class GetStagesResponse(ApiBaseModel):
    stages: Optional[List[GetStagesResponseStages]] = None
    errorCode: Optional[str] = None
    errorMessage: Optional[str] = None
    hasError: Optional[bool] = None
    httpStatusCode: Optional[HttpStatusCode] = None


import cherwell_pydantic_api.types as ct


class GetStatusesResponse(ApiBaseModel):
    statuses: Optional[List[GetStatusesResponseStatuses]] = None
    errorCode: Optional[str] = None
    errorMessage: Optional[str] = None
    hasError: Optional[bool] = None
    httpStatusCode: Optional[HttpStatusCode] = None


import cherwell_pydantic_api.types as ct


class GetTransitionsResponse(ApiBaseModel):
    transitions: Optional[List[GetTransitionsResponseTransition]] = None
    errorCode: Optional[str] = None
    errorMessage: Optional[str] = None
    hasError: Optional[bool] = None
    httpStatusCode: Optional[HttpStatusCode] = None


import cherwell_pydantic_api.types as ct


class GetTransitionOptionsResponse(ApiBaseModel):
    transitions: Optional[List[GetTransitionOptionsResponseTransition]] = None
    errorCode: Optional[str] = None
    errorMessage: Optional[str] = None
    hasError: Optional[bool] = None
    httpStatusCode: Optional[HttpStatusCode] = None



import logging
import time
from typing import Any, AsyncIterable, Iterable, Literal, Mapping, Optional, Union

from cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.BusinessObject import Summary
from cherwell_pydantic_api.settings import InstanceSettingsBase
from ._generated.api.endpoints import GeneratedInterfaces
from .generated_api_utils import GeneratedInterfaceBase
import httpx



from .types import BusObID

URLType = Union[str, httpx.URL]


class Connection(GeneratedInterfaces, GeneratedInterfaceBase):
    def __init__(self, instance_settings: InstanceSettingsBase):
        self._settings = instance_settings
        self._base_url = self._settings.base_url
        self._client_id = self._settings.client_id
        self._username = self._settings.username
        self._password = self._settings.password
        self._client = httpx.AsyncClient(base_url=str(self._base_url),
                                         headers={
                                             'Content-Type': 'application/json'},
                                         params={'locale': 'en-150'},
                                         timeout=self._settings.timeout)
        self._token: Optional[str] = None
        # TODO: come up with a better solution for these:
        self.raise_on_error_500 = False
        self.retry_on_error_401 = False
        self.retry_on_error_401_wait = 4.0
        self.reauthentication_counter = 0


    async def authenticate(self):
        # TODO: support other grant_type values
        response = await self.Token(grant_type='password', client_id=self._client_id, username=self._username, password=self._password)
        self._token = response.access_token
        self._client.headers.update({'Authorization': f'Bearer {self._token}'})
        return True


    async def logout(self):
        return await self.delete('/api/v1/logout')


    async def request(self,
                      method: Literal['GET', 'POST', 'PUT', 'DELETE'],
                      url: URLType,
                      *,
                      content: Optional[Union[str, bytes,
                                              Iterable[bytes], AsyncIterable[bytes]]] = None,
                      data: Optional[Mapping[str, Any]] = None,
                      json: Optional[Any] = None,
                      params: Optional[Mapping[str, Any]] = None,
                      content_type: Optional[str] = None,
                      **kwargs
                      ) -> Any:
        if content_type is not None:
            kwargs.setdefault('headers', {})['Content-Type'] = content_type
        req = self._client.build_request(
            method, url=url, content=content, data=data, json=json, params=params, **kwargs)
        self.last_request = req
        response = await self._client.send(req)
        self.last_response = response
        # TODO: come up with a better solution for this
        for i in range(5):
            if self.retry_on_error_401 and response.status_code == 401:
                logging.debug('401, retry {0}'.format(i + 1))
                time.sleep(self.retry_on_error_401_wait * i)
                await self.authenticate()
                self.reauthentication_counter += 1
                response = await self._client.request(method, url=url, data=data, json=json, params=params, **kwargs)
                self.last_response = response
        j = response.json()
        if not response.is_success and j is not None and j.get('hasError', None):
            # The Cherwell API returns a 500 status code if e.g. a business object ID is invalid, but the request was otherwise OK.
            # In this case the body contains a JSON object with a "hasError" key.
            j['httpStatusCode'] = response.status_code
            return j
        if not self.raise_on_error_500 and response.status_code == 500:
            return None
        response.raise_for_status()
        return j


    async def get(self, url: URLType, *, params: Optional[Mapping[str, Any]] = None) -> Any:
        return await self.request('GET', url, params=params)


    async def post_body(self, url: URLType, *, content: Union[str, bytes, Iterable[bytes], AsyncIterable[bytes]], content_type: str = 'application/json', params: Optional[Mapping[str, Any]] = None) -> Any:
        return await self.request('POST', url, content=content, content_type=content_type, params=params)


    async def post_form(self, url: URLType, *, data: Optional[Mapping[str, Any]] = None, params: Optional[Mapping[str, Any]] = None) -> Any:
        return await self.request('POST', url, data=data, params=params)


    async def post(self, url: URLType) -> Any:
        return await self.request('POST', url)


    async def delete(self, url, **kwargs) -> Any:
        return await self.request('DELETE', url, **kwargs)


    async def get_bo_summary(self, *, busobid: Optional[BusObID] = None, busobname: Optional[str] = None) -> Optional[Summary]:
        if busobid is not None:
            response = await self.GetBusinessObjectSummaryByIdV1(busobid=busobid)
        elif busobname is None:
            raise TypeError(
                'get_bo_summary() missing 1 required argument: busobid or busobname')
        else:
            response = await self.GetBusinessObjectSummaryByNameV1(busobname=busobname)
        if len(response) == 0:
            return None
        assert len(response) == 1
        return response[0]


    async def get_busobid(self, busobname: str) -> Optional[BusObID]:
        summary = await self.get_bo_summary(busobname=busobname)
        if summary is None:
            return None
        return summary.busObId


    # TODO: legacy, remove
    bo_name_to_id = get_busobid


    def get_bo(self, publicId, busObId=None, busObName=None):
        if busObId is None:
            busObId = self.bo_name_to_id(busObName)
        r = self.get(
            '/api/v1/getbusinessobject/busobid/{0}/publicid/{1}'.format(busObId, publicId))
        return r


    def get_related_bo(self, parentBusObId, parentBusObRecId, relationshipId):
        r = self.get('/api/v1/getrelatedbusinessobject/parentbusobid/{0}/parentbusobrecid/{1}/relationshipid/{2}'.format(
            parentBusObId, parentBusObRecId, relationshipId))
        return r


    def save_bo(self, bo_dict):
        req = bo_dict.save_change()
        self.last_save_bo_req = req
        if req is None:
            return None
        r = self.post('/api/v1/savebusinessobject', req)
        if r is None:
            return None
        bo_dict.clear_dirty()
        if bo_dict.busObRecId is None:
            bo_dict.busObRecId = r['busObRecId']
        if bo_dict.busObPublicId is None:
            bo_dict.busObPublicId = r['busObPublicId']
        return r


    def link_related_bo(self, parentBusObId, parentBusObRecId, relationshipId, busObId, busObRecId):
        r = self.get('/api/v1/linkrelatedbusinessobject/parentbusobid/{0}/parentbusobrecid/{1}/relationshipid/{2}/busobid/{3}/busobrecid/{4}'.format(
            parentBusObId, parentBusObRecId, relationshipId, busObId, busObRecId))
        return r


    def get_search_results_operator(self, bo_dict, op_dict, fields=None):
        change = bo_dict.save_change()
        search = {'busObId': change['busObId']}
        if fields is not None:
            field_ids = []
            for field in fields:
                if 'BO:' in field:
                    field_ids.append(field)
                else:
                    field_ids.append(bo_dict.fieldIds[field])
            search['fields'] = field_ids
        filters = []
        search['filters'] = filters
        for field in change['fields']:
            op = op_dict.get(field['name'], op_dict.get(
                field['fieldId'], 'eq'))
            filters.append(
                {'fieldId': field['fieldId'], 'value': field['value'], 'operator': op})
        search['pageSize'] = 100000
        self.last_search = search
        r = self.post('/api/v1/getsearchresults', search)
        return r


    def get_search_results_eq(self, bo_dict, fields=None):
        return self.get_search_results_operator(bo_dict=bo_dict, op_dict={}, fields=fields)




from pathlib import Path
from typing import Any, Optional

import black
from jinja2 import Environment, PackageLoader, select_autoescape

from cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.BusinessObject import FieldDefinition
from cherwell_pydantic_api.bo.valid_schema import ValidSchema
from cherwell_pydantic_api.settings import InstanceSettingsBase, Settings



class ParsedFieldDefinition(FieldDefinition):
    python_type: str = 'Any'
    python_type_module: Optional[str] = None
    pydantic_field_args: dict[str, Any] = {}
    fieldId: str  # override Optional[str] with str

    @property
    def pydantic_field_params(self) -> str:
        return ', '.join([f'{k}={v}' for k, v in self.pydantic_field_args.items()])

    @property
    def fieldid_parts(self) -> dict[str, str]:
        """The fieldId has the format: 'BO:947dfef9ae6b072ca567a444dca79d9f9ea47a112f,FI:9492462c89a8f7785f1537412ca51a2f218293edb0,RE:9492535dfbf4286b49047f45bab2f201737e739a46'
        This method returns a dict with the keys 'BO', 'FI', 'RE', etc. and the values are the corresponding IDs."""
        r = {}
        for part in self.fieldId.split(','):
            k, v = part.split(':', 1)
            r[k.upper()] = v.lower()
        return r

    def resolve_type(self):
        if self.type == 'Text':
            self.python_type = 'str'
            if self.maximumSize:
                try:
                    self.pydantic_field_args['max_length'] = int(
                        self.maximumSize)
                except:
                    pass
        elif self.type == 'DateTime':
            self.python_type_module = 'datetime'
            if self.hasDate and self.hasTime:
                self.python_type = 'datetime.datetime'
            elif self.hasDate:
                self.python_type = 'datetime.date'
            elif self.hasTime:
                self.python_type = 'datetime.time'
            else:
                self.python_type = 'datetime.datetime'
        elif self.type == 'Number':
            if self.decimalDigits == 0:
                self.python_type = 'int'
            else:
                self.python_type = 'decimal.Decimal'
                self.python_type_module = 'decimal'
                self.pydantic_field_args['decimal_places'] = self.decimalDigits
                if self.wholeDigits:
                    self.pydantic_field_args['max_digits'] = self.wholeDigits
        elif self.type == 'Logical':
            self.python_type = 'bool'
        else:
            raise ValueError(
                f"Unknown type: {self.type} for field {self.name}")
        if not self.required:
            self.python_type = f"Optional[{self.python_type}]"



class PydanticModelGenerator:
    def __init__(self, instance_settings: InstanceSettingsBase):
        self._instance_settings = instance_settings
        self._jinja_env = Environment(loader=PackageLoader('cherwell_pydantic_api.bo', 'templates'),
                                      autoescape=select_autoescape(['py']), trim_blocks=True, lstrip_blocks=True)

    def generate_model(self, schema: ValidSchema) -> str:
        template = self._jinja_env.get_template('model.py.j2')
        fields = [ParsedFieldDefinition(**field.dict())
                  for field in schema.fieldDefinitions]
        modules = set()
        for field in fields:
            field.resolve_type()
            if field.python_type_module:
                modules.add(field.python_type_module)
            assert field.fieldid_parts['BO'] == schema.busObId
        return template.render(schema=schema, fields=fields, modules=modules, settings=self._instance_settings)


    def generate_model_file(self, schema: ValidSchema) -> Path:
        model_str = self.generate_model(schema)
        model_dir = Settings().model_dir / self._instance_settings.name
        if not model_dir.exists():
            model_dir.mkdir()
            (model_dir / '__init__.py').touch()
        model_file = model_dir / f'{schema.name.lower()}.py'
        model_file.write_text(black.format_str(
            model_str, mode=black.FileMode(preview=True)))

        return model_file



# generated by datamodel-codegen:
#   filename:  swagger.json
#   timestamp: 2023-03-19T19:24:29+00:00

from __future__ import annotations

from enum import Enum
from typing import List, Optional

import cherwell_pydantic_api.types as ct
from cherwell_pydantic_api.generated_api_utils import ApiBaseModel, HttpStatusCode

from .... import CherwellObjectID
from . import BusinessObject, Core


class SectionField(ApiBaseModel):
    attributes: Optional[List[CherwellObjectID]] = None
    fieldId: Optional[str] = None
    fieldType: Optional[str] = None
    label: Optional[str] = None
    multiline: Optional[bool] = None
    value: Optional[str] = None


import cherwell_pydantic_api.types as ct


class Section(ApiBaseModel):
    sectionFields: Optional[List[SectionField]] = None
    galleryImage: Optional[str] = None
    title: Optional[str] = None
    relationshipId: Optional[str] = None
    targetBusObId: Optional[ct.BusObID] = None
    targetBusObRecId: Optional[ct.BusObRecID] = None


import cherwell_pydantic_api.types as ct


class MobileFormResponse(ApiBaseModel):
    actions: Optional[List[Core.Action]] = None
    attachments: Optional[List[BusinessObject.Attachment]] = None
    galleryImage: Optional[str] = None
    locationInformation: Optional[Core.Location] = None
    sections: Optional[List[Section]] = None
    title: Optional[str] = None
    errorCode: Optional[str] = None
    errorMessage: Optional[str] = None
    hasError: Optional[bool] = None
    httpStatusCode: Optional[HttpStatusCode] = None



#!/usr/bin/env python3
"""Cherwell API code generator

Generates Python models and endpoints from the Cherwell API Swagger schema.
This is NOT intended as a general purpose code generator. It will be tested only
against the Cherwell Swagger schema and may not work with other schemas.
"""


import re
from collections import defaultdict
from pathlib import Path
from typing import Dict, Set, Tuple

import black
import datamodel_code_generator
import datamodel_code_generator.format



re_fstr = re.compile(r'{([^}]+)}')


def class_name_generator(name: str) -> str:
    if name == 'Object':
        return 'CherwellObjectID'
    if name in ('Model', 'Object', 'NameValuePair', 'Link'):
        return f"Cherwell{name}"
    return name


def generate_models(input: Path, output: Path):
    # I had to patch datamodel_code_generator, will submit a PR. auk 20230316
    # PR: https://github.com/koxudaxi/datamodel-code-generator/pull/1187
    return datamodel_code_generator.generate(
        input,
        input_file_type=datamodel_code_generator.InputFileType.JsonSchema,
        output=output,
        target_python_version=datamodel_code_generator.format.PythonVersion.PY_39,
        base_class='cherwell_pydantic_api.generated_api_utils.ApiBaseModel',
        # Required so that the __root__ field is included - required for pydantic schema
        custom_template_dir=Path(
            __file__).parent.joinpath('codegen_templates'),
        field_include_all_keys=True,
        custom_class_name_generator=class_name_generator
    )


def get_endpoint_method(path: str, op: str, opspec: Dict) -> Tuple[str, str]:
    opid = opspec['operationId']
    tag = opspec['tags'][0]
    if opid.startswith(f"{tag}_"):
        opid = opid[len(f"{tag}_"):]
    return (tag, opid)


def get_python_type(paramspec: Dict, opspec: Dict, models_module: str, import_set: Set) -> str:
    if not paramspec.get('required', True):
        optional_paramspec = paramspec.copy()
        del optional_paramspec['required']
        return f"Optional[{get_python_type(optional_paramspec, opspec, models_module, import_set)}]"
    if '$ref' in paramspec:
        ref = paramspec['$ref']
        if ref.startswith('#/definitions/'):
            model = models_module + '.' + ref[len('#/definitions/'):]
            import_set.add('.'.join(model.split('.')[:-1]))
            return model
    if 'schema' in paramspec:
        schema = paramspec['schema']
        if schema.get('type', None) == 'array':
            return f"List[{get_python_type(schema['items'], opspec, models_module, import_set)}]"
        return get_python_type(schema, opspec, models_module, import_set)
    name = paramspec.get('name', '').lower()
    is_input = paramspec.get('in', '') in ('query', 'path',)
    if name.endswith('busobid'):
        import_set.add('cherwell_pydantic_api.types')
        return 'cherwell_pydantic_api.types.BusObIDParamType' if is_input else 'cherwell_pydantic_api.types.BusObID'
    if name.endswith('busobrecid'):
        import_set.add('cherwell_pydantic_api.types')
        return 'cherwell_pydantic_api.types.BusObRecID'
    enumvals = paramspec.get('enum', None)
    if enumvals:
        return (f"Literal{enumvals}")
    if 'application/octet-stream' in opspec['consumes'] and paramspec['type'] == 'string':
        paramspec['pytype'] = 'Union[str, bytes, Iterable[bytes], AsyncIterable[bytes]]'
    if opspec['operationId'] == 'Service_Token' and name in ('client_id', 'client_secret', 'password',):
        import_set.add('pydantic')
        return 'pydantic.SecretStr'
    if 'type' in paramspec:
        return {'boolean': 'bool', 'integer': 'int', 'number': 'float', 'string': 'str',
                'array': 'List', 'object': 'Dict'}.get(paramspec['type'], 'Any')
    return 'Any'


def format_docstring(s: str) -> str:
    return re.sub('(?:<br */?>)|(?:</br *>)', '\n', s).replace('\\', '&#92;')


def generate_endpoints(output: Path):
    try:
        import cherwell_pydantic_api._generated.api.models
        schema = cherwell_pydantic_api._generated.api.models.CherwellModel.schema()
    except:
        del cherwell_pydantic_api._generated.api.models  # type: ignore
        raise
    models_module = cherwell_pydantic_api._generated.api.models.__name__
    tags = {}
    for tag in schema['tags']:
        tags[tag['name']] = tag['description']
    interfaces = defaultdict(lambda: [])
    interface_imports = defaultdict(lambda: set())
    for path, pathspec in schema['paths'].items():
        for op, opspec in pathspec.items():
            assert op in ('get', 'post', 'put', 'delete')
            interface, method = get_endpoint_method(path, op, opspec)
            if method == '':
                continue

            # Parse the parameters

            paramspecs = []
            bodyparam = None
            for paramspec in opspec['parameters']:
                # Skip lang and locale parameters. TODO: consider offering these as parameters
                if paramspec['name'] in ('lang', 'locale',):
                    continue
                pytype = get_python_type(
                    paramspec, opspec, models_module, interface_imports[interface])
                paramspec['pytype'] = pytype
                paramspec['pyvalue'] = paramspec['name']
                if 'SecretStr' in pytype:
                    paramspec['pyvalue'] = f"{paramspec['name']}.get_secret_value()"
                if paramspec['in'] == 'body':
                    assert bodyparam is None
                    bodyparam = paramspec
                paramspecs.append(paramspec)

            # Determine the return type

            if '200' in opspec['responses']:
                rtype = get_python_type(
                    opspec['responses']['200'], opspec, models_module, interface_imports[interface])
            else:
                rtype = 'None'

            # Generate method signature

            if paramspecs:
                ilines = [f" async def {method}(self,"]
                for paramspec in paramspecs:
                    if paramspec['required']:
                        ilines.append(
                            f"  {paramspec['name']}: {paramspec['pytype']},")
                    else:
                        ilines.append(
                            f"  {paramspec['name']}: {paramspec['pytype']} = None,")
                ilines.append(f" ) -> {rtype}:")
            else:
                ilines = [f" async def {method}(self) -> {rtype}:"]

            # Generate method docstring

            if 'summary' in opspec:
                ilines.append(f"  \"\"\"{format_docstring(opspec['summary'])}\n")
            else:
                ilines.append('  """')
            if 'description' in opspec:
                ilines.append(format_docstring(opspec['description']))
            for paramspec in paramspecs:
                description = paramspec.get('description', '')
                if description == '' and paramspec['in'] == 'body':
                    description = 'The request body'
                description = format_docstring(description)
                ilines.append(
                    f" :param {paramspec['name']}: {description}")
            ilines.append(f" :return: {rtype}")
            ilines.append('"""')

            # Generate params to pass to the request

            qparams = []
            qparams_optional = []
            post = []
            post_optional = []
            pathparams = []
            path_fstrs = set(re_fstr.findall(path))

            for paramspec in paramspecs:
                if paramspec['in'] == 'query':
                    if not paramspec['required']:
                        qparams_optional.append(
                            f"  if {paramspec['name']} is not None: params['{paramspec['name']}'] = {paramspec['pyvalue']}")
                    else:
                        qparams.append(
                            f"'{paramspec['name']}': {paramspec['pyvalue']}")
                elif paramspec['in'] == 'formData':
                    assert op in ('post', 'post_form',)
                    assert bodyparam is None
                    op = 'post_form'
                    if not paramspec['required']:
                        post_optional.append(
                            f"  if {paramspec['name']} is not None: post['{paramspec['name']}'] = {paramspec['pyvalue']}")
                    else:
                        post.append(
                            f"'{paramspec['name']}': {paramspec['pyvalue']}")
                elif paramspec['in'] == 'path':
                    assert (paramspec['required'])
                    pathparams.append(paramspec['name'])
                    if not paramspec['name'].lower().endswith('id'):
                        print(f"pathparam not *id: {paramspec['name']}")
                    if paramspec['pytype'] not in ('int', 'float'):
                        ilines.append(
                            f"  self.validate_path_param({paramspec['name']})")
                    path_fstrs.remove(paramspec['name'])
                elif paramspec is bodyparam:
                    assert op in ('post', 'put',)
                    op = 'post_body'
                else:
                    assert False, f"Unknown param type {paramspec['in']} (path: {path})"
            assert not (
                pathparams and path_fstrs), f"Unused path params: {path_fstrs}"

            # Generate params used to call request

            call_params = [f"{'f' if pathparams else ''}'{path}'"]
            if post:
                ilines.append(f"  post={{{', '.join(post)}}}")
            elif post_optional:
                ilines.append("  post={}")
            if qparams:
                ilines.append(f"  params: dict[str, Any]={{{', '.join(qparams)}}}")
                call_params.append('params=params')
            elif qparams_optional:
                ilines.append("  params={}")
                call_params.append('params=params')
            if post_optional:
                ilines.extend(post_optional)
            if qparams_optional:
                ilines.extend(qparams_optional)
            if op == 'post_form':
                call_params.append('data=post')
            elif op == 'post_body' and bodyparam:
                if 'application/json' in opspec['consumes']:
                    call_params.append(
                        f"content={bodyparam['pyvalue']}.json(exclude_unset=True, by_alias=True)")
                elif 'application/octet-stream' in opspec['consumes']:
                    call_params.append(f"content={bodyparam['pyvalue']}")
                    call_params.append(
                        "content_type='application/octet-stream'")
                else:
                    assert False, "Unknown content type"

            # Call the API

            if rtype == 'None':
                ilines.append(f"  await self.{op}({', '.join(call_params)})")
                ilines.append('  return None')
            elif '.' not in rtype:
                ilines.append(
                    f"  return await self.{op}({', '.join(call_params)})")
            else:
                ilines.append(
                    f"  response = await self.{op}({', '.join(call_params)})")
                ilines.append(f"  return parse_obj_as({rtype}, response)")
            ilines.append('')

            # End of method

            interfaces[interface].extend(ilines)
    ifiles = {}
    init_lines = []
    for interface in interfaces.keys():
        ilines = interfaces[interface]
        hlines = ["from typing import Any, Dict, List, Literal, Optional, Union, AsyncIterable, Iterable",
                  "from pydantic import parse_obj_as"]
        hlines.extend([f"import {i}" for i in interface_imports[interface]])
        hlines.append(
            "from cherwell_pydantic_api.generated_api_utils import GeneratedInterfaceBase\n")
        hlines.append(f"class {interface}Interface(GeneratedInterfaceBase):")
        ifiles[interface] = '\n'.join(hlines + ilines)
        init_lines.append(f"from .{interface} import {interface}Interface")
    init_lines.append(
        f"\n\nclass GeneratedInterfaces({', '.join([f'{i}Interface' for i in interfaces.keys()])}):\n pass")
    ifiles['__init__'] = '\n'.join(init_lines)
    if not output.exists():
        output.mkdir(parents=True)
    for interface in ifiles.keys():
        file = output.joinpath(f"{interface}.py").open('wt')
        file.write(black.format_str(
            ifiles[interface], mode=black.FileMode(preview=True)))
        file.close()
    return ifiles


def generate(base: Path):
    apipath = base.joinpath('cherwell_pydantic_api/_generated/api')
    if not base.joinpath('cherwell_pydantic_api/__init__.py').exists():
        raise Exception(
            f"Run this from the root of the cherwell_pydantic_api project")
    if not apipath.exists():
        apipath.mkdir(parents=True)
        apipath.joinpath('__init__.py').touch()
    try:
        r2 = generate_endpoints(apipath.joinpath('endpoints'))
        r1 = None
        print("Models were recycled")
    except AttributeError as e:
        print(
            f"Failed to generate endpoints, trying to regenerate models; {e}")
        r1 = generate_models(base.joinpath('swagger.json'),
                             apipath.joinpath('models'))
        r2 = generate_endpoints(apipath.joinpath('endpoints'))
    return r1, r2


if __name__ == '__main__':
    r1, r2 = generate(Path('.'))



from typing import Any, Dict, List, Literal, Optional, Union, AsyncIterable, Iterable
from pydantic import parse_obj_as
import cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts
import cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.BusinessObject
import cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.Searches
import cherwell_pydantic_api.types
from cherwell_pydantic_api.generated_api_utils import GeneratedInterfaceBase


class BusinessObjectInterface(GeneratedInterfaceBase):
    async def DeleteBusinessObjectBatchV1(
        self,
        request: cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.BusinessObject.BatchDeleteRequest,
    ) -> cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.BusinessObject.BatchDeleteResponse:
        """Delete Business Objects in a batch

        Operation to delete a batch of Business Objects.
         :param request: Specify an array of Business Object IDs and record IDs or public IDs. Use a flag to stop on error or continue on error.
         :return: cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.BusinessObject.BatchDeleteResponse
        """
        response = await self.post_body(
            "/api/V1/deletebusinessobjectbatch",
            content=request.json(exclude_unset=True, by_alias=True),
        )
        return parse_obj_as(
            cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.BusinessObject.BatchDeleteResponse,
            response,
        )

    async def DeleteBusinessObjectByPublicIdV1(
        self,
        busobid: cherwell_pydantic_api.types.BusObIDParamType,
        publicid: str,
    ) -> cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.BusinessObject.DeleteResponse:
        """Delete a Business Object by public ID

        Operation to delete a Business Object by Business Object ID.
         :param busobid: Specify the Business Object ID.
         :param publicid: Specify the Business Object public ID.
         :return: cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.BusinessObject.DeleteResponse
        """
        self.validate_path_param(busobid)
        self.validate_path_param(publicid)
        response = await self.delete(
            f"/api/V1/deletebusinessobject/busobid/{busobid}/publicid/{publicid}"
        )
        return parse_obj_as(
            cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.BusinessObject.DeleteResponse,
            response,
        )

    async def DeleteBusinessObjectByRecIdV1(
        self,
        busobid: cherwell_pydantic_api.types.BusObIDParamType,
        busobrecid: cherwell_pydantic_api.types.BusObRecID,
    ) -> cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.BusinessObject.DeleteResponse:
        """Delete a Business Object by record ID

        Operation to delete a single Business Object.
         :param busobid: Specify the Business Object ID.
         :param busobrecid:  Specify the Business Object record ID.
         :return: cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.BusinessObject.DeleteResponse
        """
        self.validate_path_param(busobid)
        self.validate_path_param(busobrecid)
        response = await self.delete(
            f"/api/V1/deletebusinessobject/busobid/{busobid}/busobrecid/{busobrecid}"
        )
        return parse_obj_as(
            cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.BusinessObject.DeleteResponse,
            response,
        )

    async def DeleteRelatedBusinessObjectByPublicIdV1(
        self,
        parentbusobid: cherwell_pydantic_api.types.BusObIDParamType,
        parentbusobrecid: cherwell_pydantic_api.types.BusObRecID,
        relationshipid: str,
        publicid: str,
    ) -> cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.Searches.RelatedBusinessObjectResponse:
        """Delete a related Business Object by public ID

        Operation to delete a related Business Object. (Use "Unlink Related Business Object" to unlink two Business Objects rather that deleting the related Business Object.)
         :param parentbusobid: Specify the Business Object ID for the parent Business Object.
         :param parentbusobrecid: Specify the record ID for the parent Business Object
         :param relationshipid: Specify the Relationship ID for the related Business Object you want to delete.
         :param publicid: Specify the public ID for the related Business Object you want to delete. Use only for Business Objects with unique public IDs. Use "Delete a related Business Object by record ID" when public IDs are not unique.
         :return: cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.Searches.RelatedBusinessObjectResponse
        """
        self.validate_path_param(parentbusobid)
        self.validate_path_param(parentbusobrecid)
        self.validate_path_param(relationshipid)
        self.validate_path_param(publicid)
        response = await self.delete(
            f"/api/V1/deleterelatedbusinessobject/parentbusobid/{parentbusobid}/parentbusobrecid/{parentbusobrecid}/relationshipid/{relationshipid}/publicid/{publicid}"
        )
        return parse_obj_as(
            cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.Searches.RelatedBusinessObjectResponse,
            response,
        )

    async def DeleteRelatedBusinessObjectByRecIdV1(
        self,
        parentbusobid: cherwell_pydantic_api.types.BusObIDParamType,
        parentbusobrecid: cherwell_pydantic_api.types.BusObRecID,
        relationshipid: str,
        busobrecid: cherwell_pydantic_api.types.BusObRecID,
    ) -> cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.Searches.RelatedBusinessObjectResponse:
        """Delete a related Business Object by record ID

        Operation to delete a related Business Object. (Use "Unlink Related Business Object" to unlink two Business Objects rather that deleting the related Business Object.)
         :param parentbusobid: Specify the Business Object ID for the parent Business Object.
         :param parentbusobrecid: Specify the record ID for the parent Business Object
         :param relationshipid: Specify the Relationship ID for the related Business Object you want to delete.
         :param busobrecid: Specify the record ID for the related Business Object you want to delete.
         :return: cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.Searches.RelatedBusinessObjectResponse
        """
        self.validate_path_param(parentbusobid)
        self.validate_path_param(parentbusobrecid)
        self.validate_path_param(relationshipid)
        self.validate_path_param(busobrecid)
        response = await self.delete(
            f"/api/V1/deleterelatedbusinessobject/parentbusobid/{parentbusobid}/parentbusobrecid/{parentbusobrecid}/relationshipid/{relationshipid}/busobrecid/{busobrecid}"
        )
        return parse_obj_as(
            cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.Searches.RelatedBusinessObjectResponse,
            response,
        )

    async def FieldValuesLookupV1(
        self,
        request: cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.BusinessObject.FieldValuesLookupRequest,
    ) -> cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.BusinessObject.FieldValuesLookupResponse:
        """Get lookup values for fields

        Operation to get potentially valid values for Business Object fields.
         :param request: Request object that specifies the Business Object and fields for which values are to be returned.
         :return: cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.BusinessObject.FieldValuesLookupResponse
        """
        response = await self.post_body(
            "/api/V1/fieldvalueslookup",
            content=request.json(exclude_unset=True, by_alias=True),
        )
        return parse_obj_as(
            cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.BusinessObject.FieldValuesLookupResponse,
            response,
        )

    async def GetActivitiesV1(
        self,
        busobid: cherwell_pydantic_api.types.BusObIDParamType,
        busobrecid: cherwell_pydantic_api.types.BusObRecID,
        pageSize: int,
        pageNumber: Optional[int] = None,
        activityType: Optional[
            Literal["All", "Audit", "Communication", "Notes", "Pinned"]
        ] = None,
    ) -> List[
        cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.BusinessObject.BusObActivity
    ]:
        """Retrieve all activities for a business object

        Operation to retrieve all activities for a business object. Activities are mapped to history tracking business objects.
         :param busobid: Specify the Business Object ID.
         :param busobrecid: Specify the Business Object record ID.
         :param pageSize: Specify the number of rows to return per page. Maximum value is 2000 per page.
         :param pageNumber: Specify the page number of the result set to return.
         :param activityType: The category of activities to retrieve. Will default to All if not specified.
         :return: List[cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.BusinessObject.BusObActivity]
        """
        self.validate_path_param(busobid)
        self.validate_path_param(busobrecid)
        params = {}
        if pageNumber is not None:
            params["pageNumber"] = pageNumber
        if activityType is not None:
            params["activityType"] = activityType
        response = await self.get(
            f"/api/V1/getactivities/busobid/{busobid}/busobrecid/{busobrecid}/pagesize/{pageSize}",
            params=params,
        )
        return parse_obj_as(
            List[
                cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.BusinessObject.BusObActivity
            ],
            response,
        )

    async def GetBusinessObjectAttachmentByAttachmentIdV1(
        self,
        attachmentid: str,
        busobid: cherwell_pydantic_api.types.BusObIDParamType,
        busobrecid: cherwell_pydantic_api.types.BusObRecID,
    ) -> str:
        """Get an imported Business Object attachment

        Operation to get a Business Object attachment that has been imported into the system.  HTTP Range Header can be used but is optional.
         :param attachmentid: Specify the internal ID of the attachment record that contains information about the imported file.
         :param busobid: Specify the Business Object ID.
         :param busobrecid: Specify the Business Object record ID.
         :return: str"""
        self.validate_path_param(attachmentid)
        self.validate_path_param(busobid)
        self.validate_path_param(busobrecid)
        return await self.get(
            f"/api/V1/getbusinessobjectattachment/attachmentid/{attachmentid}/busobid/{busobid}/busobrecid/{busobrecid}"
        )

    async def GetBusinessObjectAttachmentsByIdAndPublicIdV1(
        self,
        busobid: cherwell_pydantic_api.types.BusObIDParamType,
        publicid: str,
        type: Literal[
            "None",
            "File",
            "FileManagerFile",
            "BusOb",
            "History",
            "Other",
            "Solution",
            "UsedAsSolution",
            "ExternalSolution",
        ],
        attachmenttype: Literal["Imported", "Linked", "URL"],
        includelinks: Optional[bool] = None,
    ) -> cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.BusinessObject.AttachmentsResponse:
        """Get attachments by Business Object public ID

        Operation to get attachments for a Business Object by Business Object ID and public ID.
         :param busobid: Specify the Business Object ID.
         :param publicid: Specify the Business Object public ID for the record that contains the attachments.
         :param type: Record attachment type:

        None - Not applicable to the REST API.

        File - Linked files.

        FileManagerFile - Imported files.

        BusOb - Attached Business Objects.

        History - Information about the attachment, if any is available. For example, an e-mail message may store the name of an attachment sent.

        Other - Not applicable to the REST API.

         :param attachmenttype: For file types, select the type of attachment:

        Imported - Attachment was imported into database.

        Linked - Attachment is linked to an external file.

        URL - Attachment is a URL.
         :param includelinks: Flag to include hyperlinks in results. Default is false.
         :return: cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.BusinessObject.AttachmentsResponse
        """
        self.validate_path_param(busobid)
        self.validate_path_param(publicid)
        self.validate_path_param(type)
        self.validate_path_param(attachmenttype)
        params = {}
        if includelinks is not None:
            params["includelinks"] = includelinks
        response = await self.get(
            f"/api/V1/getbusinessobjectattachments/busobid/{busobid}/publicid/{publicid}/type/{type}/attachmenttype/{attachmenttype}",
            params=params,
        )
        return parse_obj_as(
            cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.BusinessObject.AttachmentsResponse,
            response,
        )

    async def GetBusinessObjectAttachmentsByIdAndRecIdV1(
        self,
        busobid: cherwell_pydantic_api.types.BusObIDParamType,
        busobrecid: cherwell_pydantic_api.types.BusObRecID,
        type: Literal[
            "None",
            "File",
            "FileManagerFile",
            "BusOb",
            "History",
            "Other",
            "Solution",
            "UsedAsSolution",
            "ExternalSolution",
        ],
        attachmenttype: Literal["Imported", "Linked", "URL"],
        includelinks: Optional[bool] = None,
    ) -> cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.BusinessObject.AttachmentsResponse:
        """Get attachments by Business Object record ID

        Operation to get attachments for a Business Object by Business Object ID and record ID.
         :param busobid: Specify the Business Object ID.
         :param busobrecid: Specify the Business Object record ID.
         :param type: Record attachment type:

        None - Not applicable to the REST API.

        File - Linked files.

        FileManagerFile - Imported files.

        BusOb - Attached Business Objects.

        History - Information about the attachment, if any is available. For example, an e-mail message may store the name of an attachment sent.

        Other - Not applicable to the REST API.

         :param attachmenttype: For file types, select the type of attachment:

        Imported - Attachment was imported into database.

        Linked - Attachment is linked to an external file.

        URL - Attachment is a URL.
         :param includelinks: Flag to include hyperlinks in results. Default is false.
         :return: cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.BusinessObject.AttachmentsResponse
        """
        self.validate_path_param(busobid)
        self.validate_path_param(busobrecid)
        self.validate_path_param(type)
        self.validate_path_param(attachmenttype)
        params = {}
        if includelinks is not None:
            params["includelinks"] = includelinks
        response = await self.get(
            f"/api/V1/getbusinessobjectattachments/busobid/{busobid}/busobrecid/{busobrecid}/type/{type}/attachmenttype/{attachmenttype}",
            params=params,
        )
        return parse_obj_as(
            cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.BusinessObject.AttachmentsResponse,
            response,
        )

    async def GetBusinessObjectAttachmentsByNameAndPublicIdV1(
        self,
        busobname: str,
        publicid: str,
        type: Literal[
            "None",
            "File",
            "FileManagerFile",
            "BusOb",
            "History",
            "Other",
            "Solution",
            "UsedAsSolution",
            "ExternalSolution",
        ],
        attachmenttype: Literal["Imported", "Linked", "URL"],
        includelinks: Optional[bool] = None,
    ) -> cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.BusinessObject.AttachmentsResponse:
        """Get attachments by Business Object name and public ID

        Operation to get attachments for a Business Object by Business Object Name and public ID.
         :param busobname: Specify the Business Object name.
         :param publicid: Specify the Business Object public ID for the record that contains the attachments.
         :param type: Record attachment type:

        None - Not applicable to the REST API.

        File - Linked files.

        FileManagerFile - Imported files.

        BusOb - Attached Business Objects.

        History - Information about the attachment, if any is available. For example, an e-mail message may store the name of an attachment sent.

        Other - Not applicable to the REST API.

         :param attachmenttype: For file types, select the type of attachment:

        Imported - Attachment was imported into database.

        Linked - Attachment is linked to an external file.

        URL - Attachment is a URL.
         :param includelinks: Flag to include hyperlinks in results. Default is false.
         :return: cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.BusinessObject.AttachmentsResponse
        """
        self.validate_path_param(busobname)
        self.validate_path_param(publicid)
        self.validate_path_param(type)
        self.validate_path_param(attachmenttype)
        params = {}
        if includelinks is not None:
            params["includelinks"] = includelinks
        response = await self.get(
            f"/api/V1/getbusinessobjectattachments/busobname/{busobname}/publicid/{publicid}/type/{type}/attachmenttype/{attachmenttype}",
            params=params,
        )
        return parse_obj_as(
            cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.BusinessObject.AttachmentsResponse,
            response,
        )

    async def GetBusinessObjectAttachmentsByNameAndRecIdV1(
        self,
        busobname: str,
        busobrecid: cherwell_pydantic_api.types.BusObRecID,
        type: Literal[
            "None",
            "File",
            "FileManagerFile",
            "BusOb",
            "History",
            "Other",
            "Solution",
            "UsedAsSolution",
            "ExternalSolution",
        ],
        attachmenttype: Literal["Imported", "Linked", "URL"],
        includelinks: Optional[bool] = None,
    ) -> cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.BusinessObject.AttachmentsResponse:
        """Get attachments by Business Object name and record ID

        Operation to get attachments for a Business Object by name and record ID.
         :param busobname: Specify the Business Object name.
         :param busobrecid: Specify the Business Object record ID.
         :param type: Record attachment type:

        None - Not applicable to the REST API.

        File - Linked files.

        FileManagerFile - Imported files.

        BusOb - Attached Business Objects.

        History - Information about the attachment, if any is available. For example, an e-mail message may store the name of an attachment sent.

        Other - Not applicable to the REST API.

         :param attachmenttype: For file types, select the type of attachment:

        Imported - Attachment was imported into database.

        Linked - Attachment is linked to an external file.

        URL - Attachment is a URL.
         :param includelinks: Flag to include hyperlinks in results. Default is false.
         :return: cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.BusinessObject.AttachmentsResponse
        """
        self.validate_path_param(busobname)
        self.validate_path_param(busobrecid)
        self.validate_path_param(type)
        self.validate_path_param(attachmenttype)
        params = {}
        if includelinks is not None:
            params["includelinks"] = includelinks
        response = await self.get(
            f"/api/V1/getbusinessobjectattachments/busobname/{busobname}/busobrecid/{busobrecid}/type/{type}/attachmenttype/{attachmenttype}",
            params=params,
        )
        return parse_obj_as(
            cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.BusinessObject.AttachmentsResponse,
            response,
        )

    async def GetBusinessObjectAttachmentsV1(
        self,
        attachmentsRequest: cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.BusinessObject.AttachmentsRequest,
    ) -> cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.BusinessObject.AttachmentsResponse:
        """Get Business Object attachments by request object

        Operation to get attachments for a Business Object by attachments request object.
         :param attachmentsRequest: Object with all the parameters to request an attachments list. You can also request a list of types to get more than just one type at a time.
         :return: cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.BusinessObject.AttachmentsResponse
        """
        response = await self.post_body(
            "/api/V1/getbusinessobjectattachments",
            content=attachmentsRequest.json(exclude_unset=True, by_alias=True),
        )
        return parse_obj_as(
            cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.BusinessObject.AttachmentsResponse,
            response,
        )

    async def GetBusinessObjectBatchV1(
        self,
        request: cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.BusinessObject.BatchReadRequest,
    ) -> cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.BusinessObject.BatchReadResponse:
        """Get a batch of Business Object records

        Operation that returns a batch of Business Object records that includes a list of field record IDs, display names, and values for each record.
         :param request: Specify an array of Business Object IDs, record IDs, or public IDs. Use a flag to stop on error or continue on error.
         :return: cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.BusinessObject.BatchReadResponse
        """
        response = await self.post_body(
            "/api/V1/getbusinessobjectbatch",
            content=request.json(exclude_unset=True, by_alias=True),
        )
        return parse_obj_as(
            cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.BusinessObject.BatchReadResponse,
            response,
        )

    async def GetBusinessObjectByPublicIdV1(
        self,
        busobid: cherwell_pydantic_api.types.BusObIDParamType,
        publicid: str,
    ) -> cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.BusinessObject.ReadResponse:
        """Get a Business Object record

        Operation that returns a Business Object record that includes a list of fields and their record IDs, names, and set values.
         :param busobid: Specify the Business Object ID.
         :param publicid: Specify the Business Object public ID.
         :return: cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.BusinessObject.ReadResponse
        """
        self.validate_path_param(busobid)
        self.validate_path_param(publicid)
        response = await self.get(
            f"/api/V1/getbusinessobject/busobid/{busobid}/publicid/{publicid}"
        )
        return parse_obj_as(
            cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.BusinessObject.ReadResponse,
            response,
        )

    async def GetBusinessObjectByRecIdV1(
        self,
        busobid: cherwell_pydantic_api.types.BusObIDParamType,
        busobrecid: cherwell_pydantic_api.types.BusObRecID,
    ) -> cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.BusinessObject.ReadResponse:
        """Get a Business Object record

        Operation that returns a Business Object record that includes a list of fields and their record IDs, names, and set values.
         :param busobid: Specify the Business Object ID.
         :param busobrecid: Specify the Business Object record ID.
         :return: cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.BusinessObject.ReadResponse
        """
        self.validate_path_param(busobid)
        self.validate_path_param(busobrecid)
        response = await self.get(
            f"/api/V1/getbusinessobject/busobid/{busobid}/busobrecid/{busobrecid}"
        )
        return parse_obj_as(
            cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.BusinessObject.ReadResponse,
            response,
        )

    async def GetBusinessObjectByScanCodeBusObIdV1(
        self,
        scanCode: str,
        busobid: cherwell_pydantic_api.types.BusObIDParamType,
    ) -> cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.BusinessObject.BarcodeLookupResponse:
        """Get a Business Object by its scan code and Business Object ID

        Operation to get a Business Object based on its associated scan code and Business Object ID.
         :param scanCode: The scan code for a Business Object record.
         :param busobid: The Business Object ID.
         :return: cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.BusinessObject.BarcodeLookupResponse
        """
        self.validate_path_param(scanCode)
        self.validate_path_param(busobid)
        response = await self.get(
            f"/api/V1/getbusinessobject/scancode/{scanCode}/busobid/{busobid}"
        )
        return parse_obj_as(
            cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.BusinessObject.BarcodeLookupResponse,
            response,
        )

    async def GetBusinessObjectByScanCodeBusObNameV1(
        self,
        scanCode: str,
        busobname: str,
    ) -> cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.BusinessObject.BarcodeLookupResponse:
        """Get a Business Object by its scan code and Business Object name

        Operation to get a Business Object based on its associated scan code and Business Object name.
         :param scanCode: The scan code for a Business Object record.
         :param busobname: The Business Bbject name.
         :return: cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.BusinessObject.BarcodeLookupResponse
        """
        self.validate_path_param(scanCode)
        self.validate_path_param(busobname)
        response = await self.get(
            f"/api/V1/getbusinessobject/scancode/{scanCode}/busobname/{busobname}"
        )
        return parse_obj_as(
            cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.BusinessObject.BarcodeLookupResponse,
            response,
        )

    async def GetBusinessObjectSchemaV1(
        self,
        busobId: cherwell_pydantic_api.types.BusObIDParamType,
        includerelationships: Optional[bool] = None,
    ) -> cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.BusinessObject.SchemaResponse:
        """Get a Business Object schema

        Operation that returns the schema for a Business Object and, optionally, its related Business Objects.
         :param busobId: Specify the Business Object ID.
         :param includerelationships: Flag to include schemas for related Business Object. Default is false.
         :return: cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.BusinessObject.SchemaResponse
        """
        self.validate_path_param(busobId)
        params = {}
        if includerelationships is not None:
            params["includerelationships"] = includerelationships
        response = await self.get(
            f"/api/V1/getbusinessobjectschema/busobid/{busobId}", params=params
        )
        return parse_obj_as(
            cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.BusinessObject.SchemaResponse,
            response,
        )

    async def GetBusinessObjectSummariesV1(
        self,
        type: Literal["All", "Major", "Supporting", "Lookup", "Groups"],
    ) -> List[
        cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.BusinessObject.Summary
    ]:
        """Get Business Object summaries by type

        Operation that returns a list of Business Object summaries by type (Major, Supporting, Lookup, Groups, and All).
         :param type: Use to show:
        All - All objects
        Major - Major objects only
        Supporting - Supporting objects only
        Lookup - Lookup objects only
        Groups - Groups only
         :return: List[cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.BusinessObject.Summary]
        """
        self.validate_path_param(type)
        response = await self.get(f"/api/V1/getbusinessobjectsummaries/type/{type}")
        return parse_obj_as(
            List[
                cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.BusinessObject.Summary
            ],
            response,
        )

    async def GetBusinessObjectSummaryByIdV1(
        self,
        busobid: cherwell_pydantic_api.types.BusObIDParamType,
    ) -> List[
        cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.BusinessObject.Summary
    ]:
        """Get a Business Object summary by ID

        Operation that returns a single Business Object summary by ID.
         :param busobid: Specify a Business Object ID to get its summary.
         :return: List[cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.BusinessObject.Summary]
        """
        self.validate_path_param(busobid)
        response = await self.get(f"/api/V1/getbusinessobjectsummary/busobid/{busobid}")
        return parse_obj_as(
            List[
                cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.BusinessObject.Summary
            ],
            response,
        )

    async def GetBusinessObjectSummaryByNameV1(
        self,
        busobname: str,
    ) -> List[
        cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.BusinessObject.Summary
    ]:
        """Get a Business Object summary by name

        Operation that returns a single Business Object summary by name.
         :param busobname: Specify a Business Object name to get its summary.
         :return: List[cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.BusinessObject.Summary]
        """
        self.validate_path_param(busobname)
        response = await self.get(
            f"/api/V1/getbusinessobjectsummary/busobname/{busobname}"
        )
        return parse_obj_as(
            List[
                cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.BusinessObject.Summary
            ],
            response,
        )

    async def GetBusinessObjectTemplateV1(
        self,
        request: cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.BusinessObject.TemplateRequest,
    ) -> cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.BusinessObject.TemplateResponse:
        """Get Business Object templates for create

        Operation that returns a template to create Business Objects.  The template includes placeholders for field values. You can then send the template with these values to the Business Object Save operation.
         :param request: Specify the Business Object ID. Use true to include all required fields or all fields. Specify an optional fields list by adding field names in a comma-delimited list ["field1", "field2"].
         :return: cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.BusinessObject.TemplateResponse
        """
        response = await self.post_body(
            "/api/V1/getbusinessobjecttemplate",
            content=request.json(exclude_unset=True, by_alias=True),
        )
        return parse_obj_as(
            cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.BusinessObject.TemplateResponse,
            response,
        )

    async def GetRelatedBusinessObjectByRequestV1(
        self,
        relatedBusinessObjectRequest: cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.Searches.RelatedBusinessObjectRequest,
        includelinks: Optional[bool] = None,
    ) -> cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.Searches.RelatedBusinessObjectResponse:
        """Get related Business Objects using a request object

        Operation to get related Business Objects for a specific relationship. Specify a list of fields to include in the response. The order of parameter usage and overrides is: all fields set to true overrides default overrides;  custom grid overrides field list settings.
         :param relatedBusinessObjectRequest: Request object containing all the possible parameters to get related Business Objects.
         :param includelinks: Flag to include hyperlinks in results. Default is false.
         :return: cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.Searches.RelatedBusinessObjectResponse
        """
        params = {}
        if includelinks is not None:
            params["includelinks"] = includelinks
        response = await self.post_body(
            "/api/V1/getrelatedbusinessobject",
            params=params,
            content=relatedBusinessObjectRequest.json(
                exclude_unset=True, by_alias=True
            ),
        )
        return parse_obj_as(
            cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.Searches.RelatedBusinessObjectResponse,
            response,
        )

    async def GetRelatedBusinessObjectV1(
        self,
        parentbusobid: cherwell_pydantic_api.types.BusObIDParamType,
        parentbusobrecid: cherwell_pydantic_api.types.BusObRecID,
        relationshipid: str,
        pageNumber: Optional[int] = None,
        pageSize: Optional[int] = None,
        allfields: Optional[bool] = None,
        usedefaultgrid: Optional[bool] = None,
        includelinks: Optional[bool] = None,
    ) -> cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.Searches.RelatedBusinessObjectResponse:
        """Get related Business Objects by ID

        Operation to get the related objects for a Business Object relationship specifying all fields or default grid as the field to return.
         :param parentbusobid: Specify the Business Object ID for the parent Business Object.
         :param parentbusobrecid: Specify the record ID for the parent Business Object.
         :param relationshipid: Specify the Relationship ID for the related Business Object you want to return.
         :param pageNumber: Specify the page number of the result set to return.
         :param pageSize: Specify the number of rows to return per page.
         :param allfields: Flag to include all related Business Object fields.  Default is true if not supplied.  If true, then UseDefaultGrid is not used.
         :param usedefaultgrid: Flag to trigger the use of the related Business Objects default grid for the list of fields to return.
         :param includelinks: Flag to include hyperlinks in results. Default is false.
         :return: cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.Searches.RelatedBusinessObjectResponse
        """
        self.validate_path_param(parentbusobid)
        self.validate_path_param(parentbusobrecid)
        self.validate_path_param(relationshipid)
        params = {}
        if pageNumber is not None:
            params["pageNumber"] = pageNumber
        if pageSize is not None:
            params["pageSize"] = pageSize
        if allfields is not None:
            params["allfields"] = allfields
        if usedefaultgrid is not None:
            params["usedefaultgrid"] = usedefaultgrid
        if includelinks is not None:
            params["includelinks"] = includelinks
        response = await self.get(
            f"/api/V1/getrelatedbusinessobject/parentbusobid/{parentbusobid}/parentbusobrecid/{parentbusobrecid}/relationshipid/{relationshipid}",
            params=params,
        )
        return parse_obj_as(
            cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.Searches.RelatedBusinessObjectResponse,
            response,
        )

    async def GetRelatedBusinessObjectWithCustomGridV1(
        self,
        parentbusobid: cherwell_pydantic_api.types.BusObIDParamType,
        parentbusobrecid: cherwell_pydantic_api.types.BusObRecID,
        relationshipid: str,
        gridid: str,
        pageNumber: Optional[int] = None,
        pageSize: Optional[int] = None,
        includelinks: Optional[bool] = None,
    ) -> cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.Searches.RelatedBusinessObjectResponse:
        """Get related Business Objects custom grid

        Operation to get related Business Objects for a specific relationship. Specify a custom grid ID as the fields to return.
         :param parentbusobid: Specify the Business Object ID for the parent Business Object.
         :param parentbusobrecid: Specify the record ID for the parent Business Object.
         :param relationshipid: Specify the Relationship ID for the related Business Object you want to return.
         :param gridid: Specify the ID for the custom grid that contains the field list.
         :param pageNumber: Specify the page number of the result set to return.
         :param pageSize: Specify the number of rows to return per page.
         :param includelinks: Flag to include hyperlinks in results. Default is false.
         :return: cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.Searches.RelatedBusinessObjectResponse
        """
        self.validate_path_param(parentbusobid)
        self.validate_path_param(parentbusobrecid)
        self.validate_path_param(relationshipid)
        self.validate_path_param(gridid)
        params = {}
        if pageNumber is not None:
            params["pageNumber"] = pageNumber
        if pageSize is not None:
            params["pageSize"] = pageSize
        if includelinks is not None:
            params["includelinks"] = includelinks
        response = await self.get(
            f"/api/V1/getrelatedbusinessobject/parentbusobid/{parentbusobid}/parentbusobrecid/{parentbusobrecid}/relationshipid/{relationshipid}/gridid/{gridid}",
            params=params,
        )
        return parse_obj_as(
            cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.Searches.RelatedBusinessObjectResponse,
            response,
        )

    async def LinkRelatedBusinessObjectByRecIdV1(
        self,
        parentbusobid: cherwell_pydantic_api.types.BusObIDParamType,
        parentbusobrecid: cherwell_pydantic_api.types.BusObRecID,
        relationshipid: str,
        busobid: cherwell_pydantic_api.types.BusObIDParamType,
        busobrecid: cherwell_pydantic_api.types.BusObRecID,
    ) -> cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.Searches.RelatedBusinessObjectResponse:
        """Link related Business Objects

        Operation to link related Business Objects.
         :param parentbusobid: Specify the Business Object ID for the parent Business Object.
         :param parentbusobrecid: Specify the record ID for the parent Business Object.
         :param relationshipid: Specify the Relationship ID for the related Business Object you want to link.
         :param busobid: Specify the Business Object ID of the Business Object to be linked.
         :param busobrecid: Specify the Business Object record ID of the Business Object to be linked.
         :return: cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.Searches.RelatedBusinessObjectResponse
        """
        self.validate_path_param(parentbusobid)
        self.validate_path_param(parentbusobrecid)
        self.validate_path_param(relationshipid)
        self.validate_path_param(busobid)
        self.validate_path_param(busobrecid)
        response = await self.get(
            f"/api/V1/linkrelatedbusinessobject/parentbusobid/{parentbusobid}/parentbusobrecid/{parentbusobrecid}/relationshipid/{relationshipid}/busobid/{busobid}/busobrecid/{busobrecid}"
        )
        return parse_obj_as(
            cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.Searches.RelatedBusinessObjectResponse,
            response,
        )

    async def LinkRelatedBusinessObjectByRecIdV2(
        self,
        parentbusobid: cherwell_pydantic_api.types.BusObIDParamType,
        parentbusobrecid: cherwell_pydantic_api.types.BusObRecID,
        relationshipid: str,
        busobid: cherwell_pydantic_api.types.BusObIDParamType,
        busobrecid: cherwell_pydantic_api.types.BusObRecID,
    ) -> cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.ResponseBase:
        """Link related Business Objects

        Operation to link related Business Objects.
         :param parentbusobid: Specify the Business Object ID for the parent Business Object.
         :param parentbusobrecid: Specify the record ID for the parent Business Object.
         :param relationshipid: Specify the Relationship ID for the related Business Object you want to link.
         :param busobid: Specify the Business Object ID of the Business Object to be linked.
         :param busobrecid: Specify the Business Object record ID of the Business Object to be linked.
         :return: cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.ResponseBase
        """
        self.validate_path_param(parentbusobid)
        self.validate_path_param(parentbusobrecid)
        self.validate_path_param(relationshipid)
        self.validate_path_param(busobid)
        self.validate_path_param(busobrecid)
        response = await self.get(
            f"/api/V2/linkrelatedbusinessobject/parentbusobid/{parentbusobid}/parentbusobrecid/{parentbusobrecid}/relationshipid/{relationshipid}/busobid/{busobid}/busobrecid/{busobrecid}"
        )
        return parse_obj_as(
            cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.ResponseBase,
            response,
        )

    async def RemoveBusinessObjectAttachmentByIdAndPublicIdV1(
        self,
        attachmentid: str,
        busobid: cherwell_pydantic_api.types.BusObIDParamType,
        publicid: str,
    ) -> None:
        """Remove an attachment by Business Object ID and public ID

        Operation to remove an attachment from a Business Object using the attachment record ID, Business Object ID, and Business Object public ID.
         :param attachmentid: Specify the internal ID of the attachment record.
         :param busobid: Specify the Business Object ID.
         :param publicid: Specify the Business Object public ID.
         :return: None"""
        self.validate_path_param(attachmentid)
        self.validate_path_param(busobid)
        self.validate_path_param(publicid)
        await self.delete(
            f"/api/V1/removebusinessobjectattachment/attachmentid/{attachmentid}/busobid/{busobid}/publicid/{publicid}"
        )
        return None

    async def RemoveBusinessObjectAttachmentByIdAndRecIdV1(
        self,
        attachmentid: str,
        busobid: cherwell_pydantic_api.types.BusObIDParamType,
        busobrecid: cherwell_pydantic_api.types.BusObRecID,
    ) -> None:
        """Remove an attachment by Business Object ID and record ID

        Operation to remove an attachment from a Business Object using the attachment record ID, Business Object ID, and Business Object record ID.
         :param attachmentid: Specify the internal ID of the attachment record.
         :param busobid: Specify the Business Object ID.
         :param busobrecid: Specify the Business Object record ID.
         :return: None"""
        self.validate_path_param(attachmentid)
        self.validate_path_param(busobid)
        self.validate_path_param(busobrecid)
        await self.delete(
            f"/api/V1/removebusinessobjectattachment/attachmentid/{attachmentid}/busobid/{busobid}/busobrecid/{busobrecid}"
        )
        return None

    async def RemoveBusinessObjectAttachmentByNameAndPublicIdV1(
        self,
        attachmentid: str,
        busobname: str,
        publicid: str,
    ) -> None:
        """Remove an attachment by Business Object name and public ID

        Operation to remove an attachment from a Business Object using the attachment record ID, Business Object name, and Business Object record ID.
         :param attachmentid: Specify the internal ID of the attachment record.
         :param busobname: Specify the Business Object name.
         :param publicid: Specify the Business Object public ID.
         :return: None"""
        self.validate_path_param(attachmentid)
        self.validate_path_param(busobname)
        self.validate_path_param(publicid)
        await self.delete(
            f"/api/V1/removebusinessobjectattachment/attachmentid/{attachmentid}/busobname/{busobname}/publicid/{publicid}"
        )
        return None

    async def RemoveBusinessObjectAttachmentByNameAndRecIdV1(
        self,
        attachmentid: str,
        busobname: str,
        busobrecid: cherwell_pydantic_api.types.BusObRecID,
    ) -> None:
        """Remove an attachment by Business Object name and record ID

        Operation to remove an attachment from a Business Object using the attachment record ID, Business Object name, and Business Object public ID.
         :param attachmentid: Specify the internal ID of the attachment record.
         :param busobname: Specify the Business Object name.
         :param busobrecid: Specify the Business Object record ID.
         :return: None"""
        self.validate_path_param(attachmentid)
        self.validate_path_param(busobname)
        self.validate_path_param(busobrecid)
        await self.delete(
            f"/api/V1/removebusinessobjectattachment/attachmentid/{attachmentid}/busobname/{busobname}/busobrecid/{busobrecid}"
        )
        return None

    async def SaveBusinessObjectAttachmentBusObV1(
        self,
        request: cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.BusinessObject.SaveBusObAttachmentRequest,
    ) -> cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.BusinessObject.AttachmentsResponse:
        """Attach a Business Object to a Business Object

        Operation to attach a Business Object to a Business Object. This links the Business Object but does not create a relationship between the two. (Use "Link Related Business Objects" to create a relationship.)
         :param request: Request object used to specify the Business Objects to attach. You can use Business Object name or ID and Business Object record ID or public ID.
         :return: cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.BusinessObject.AttachmentsResponse
        """
        response = await self.post_body(
            "/api/V1/savebusinessobjectattachmentbusob",
            content=request.json(exclude_unset=True, by_alias=True),
        )
        return parse_obj_as(
            cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.BusinessObject.AttachmentsResponse,
            response,
        )

    async def SaveBusinessObjectAttachmentLinkV1(
        self,
        request: cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.BusinessObject.SaveLinkAttachmentRequest,
    ) -> cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.BusinessObject.AttachmentsResponse:
        """Attach a file via UNC

        Operation to attach a file to a Business Object via a path (UNC recommended).
         :param request: Request object used to specify the file path (UNC recommended) and the Business Object. You can use Business Object name or ID and Business Object record ID or public ID.
         :return: cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.BusinessObject.AttachmentsResponse
        """
        response = await self.post_body(
            "/api/V1/savebusinessobjectattachmentlink",
            content=request.json(exclude_unset=True, by_alias=True),
        )
        return parse_obj_as(
            cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.BusinessObject.AttachmentsResponse,
            response,
        )

    async def SaveBusinessObjectAttachmentUrlV1(
        self,
        request: cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.BusinessObject.SaveUrlAttachmentRequest,
    ) -> cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.BusinessObject.AttachmentsResponse:
        """Attach a URL path

        Operation to attach a URL path to a Business Object.
         :param request: Request object used to specify the URL path and Business Object. You can use Business Object name or ID and Business Object record ID or public ID.
         :return: cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.BusinessObject.AttachmentsResponse
        """
        response = await self.post_body(
            "/api/V1/savebusinessobjectattachmenturl",
            content=request.json(exclude_unset=True, by_alias=True),
        )
        return parse_obj_as(
            cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.BusinessObject.AttachmentsResponse,
            response,
        )

    async def SaveBusinessObjectBatchV1(
        self,
        request: cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.BusinessObject.BatchSaveRequest,
    ) -> cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.BusinessObject.BatchSaveResponse:
        """Create or update a batch of Business Objects

        Operation that creates or updates an array of Business Objects in a batch. To update, specify record ID or public ID. To create, leave record ID and public ID empty.
         :param request: Specify the array of Business Object templates.
         :return: cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.BusinessObject.BatchSaveResponse
        """
        response = await self.post_body(
            "/api/V1/savebusinessobjectbatch",
            content=request.json(exclude_unset=True, by_alias=True),
        )
        return parse_obj_as(
            cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.BusinessObject.BatchSaveResponse,
            response,
        )

    async def SaveBusinessObjectV1(
        self,
        request: cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.BusinessObject.SaveRequest,
    ) -> cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.BusinessObject.SaveResponse:
        """Create or Update a Business Object

        Operation that creates a new Business Object or updates an existing Business Object. To create, leave record ID and public ID empty. Upon creating or saving, a cache key is returned to use for subsequent requests. If the object is not found in the cache with said cache key, specify record ID or public ID to save and return a new cache key. Set persist = true, to actually save the Business Object to disk, persist = false will just cache it.
         :param request: Specify a list of fields from a Business Object template.
         :return: cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.BusinessObject.SaveResponse
        """
        response = await self.post_body(
            "/api/V1/savebusinessobject",
            content=request.json(exclude_unset=True, by_alias=True),
        )
        return parse_obj_as(
            cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.BusinessObject.SaveResponse,
            response,
        )

    async def SaveRelatedBusinessObjectV1(
        self,
        request: cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.BusinessObject.RelatedSaveRequest,
    ) -> cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.BusinessObject.RelatedSaveResponse:
        """Create or update a related Business Object

        Operation that creates or updates a related Business Object. To update, specify record ID or public ID. To create, leave record ID and public ID empty.
         :param request: Request object specifying the parent the Business Object, the Relationship, and field values for the Business Object to create or update.
         :return: cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.BusinessObject.RelatedSaveResponse
        """
        response = await self.post_body(
            "/api/V1/saverelatedbusinessobject",
            content=request.json(exclude_unset=True, by_alias=True),
        )
        return parse_obj_as(
            cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.BusinessObject.RelatedSaveResponse,
            response,
        )

    async def UnLinkRelatedBusinessObjectByRecIdV1(
        self,
        parentbusobid: cherwell_pydantic_api.types.BusObIDParamType,
        parentbusobrecid: cherwell_pydantic_api.types.BusObRecID,
        relationshipid: str,
        busobid: cherwell_pydantic_api.types.BusObIDParamType,
        busobrecid: cherwell_pydantic_api.types.BusObRecID,
    ) -> cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.Searches.RelatedBusinessObjectResponse:
        """UnLink related Business Objects

        Operation to unlink related Business Objects.
         :param parentbusobid: Specify the Business Object ID for the parent Business Object.
         :param parentbusobrecid: Specify the record ID for the parent Business Object.
         :param relationshipid: Specify the Relationship ID for the related Business Object you want to unlink.
         :param busobid: Specify the Business Object ID of the Business Object to be unlinked.
         :param busobrecid: Specify the Business Object record ID of the Business Object to be unlinked.
         :return: cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.Searches.RelatedBusinessObjectResponse
        """
        self.validate_path_param(parentbusobid)
        self.validate_path_param(parentbusobrecid)
        self.validate_path_param(relationshipid)
        self.validate_path_param(busobid)
        self.validate_path_param(busobrecid)
        response = await self.delete(
            f"/api/V1/unlinkrelatedbusinessobject/parentbusobid/{parentbusobid}/parentbusobrecid/{parentbusobrecid}/relationshipid/{relationshipid}/busobid/{busobid}/busobrecid/{busobrecid}"
        )
        return parse_obj_as(
            cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.Searches.RelatedBusinessObjectResponse,
            response,
        )

    async def UploadBusinessObjectAttachmentByIdAndPublicIdV1(
        self,
        body: str,
        filename: str,
        busobid: cherwell_pydantic_api.types.BusObIDParamType,
        publicid: str,
        offset: int,
        totalsize: int,
        attachmentid: Optional[str] = None,
        displaytext: Optional[str] = None,
    ) -> str:
        """Upload an attachment by Business Object ID and public ID

        Operation to upload an attachment to a Business Object record using a Business Object ID and public ID. The body of the request is the byte array of the file part being uploaded.
         :param body: The request body
         :param filename: Specify the name of the file being uploaded. If no attachment name is provided, the file name is used.
         :param busobid: Specify the Business Object ID.
         :param publicid: Specify the Business Object public ID  to attach the file to.
         :param offset: The offset is the starting index of the file part being uploaded.  If this is the first part then the offset will be zero.
         :param totalsize: The entire file size in bytes.
         :param attachmentid: Specify the attachment ID of an uploaded file to upload subsequent parts and ensure each part gets appended to the parts that have already been uploaded.
         :param displaytext: Specify the attachment name, which is the display text for the attachment record.
         :return: str"""
        self.validate_path_param(filename)
        self.validate_path_param(busobid)
        self.validate_path_param(publicid)
        params = {}
        if attachmentid is not None:
            params["attachmentid"] = attachmentid
        if displaytext is not None:
            params["displaytext"] = displaytext
        return await self.post_body(
            f"/api/V1/uploadbusinessobjectattachment/filename/{filename}/busobid/{busobid}/publicid/{publicid}/offset/{offset}/totalsize/{totalsize}",
            params=params,
            content=body,
            content_type="application/octet-stream",
        )

    async def UploadBusinessObjectAttachmentByIdAndRecIdV1(
        self,
        body: str,
        filename: str,
        busobid: cherwell_pydantic_api.types.BusObIDParamType,
        busobrecid: cherwell_pydantic_api.types.BusObRecID,
        offset: int,
        totalsize: int,
        attachmentid: Optional[str] = None,
        displaytext: Optional[str] = None,
    ) -> str:
        """Upload an attachment by Business Object ID and record ID

        Operation to upload an attachment to a Business Object record using a Business Object ID and record ID. The body of the request is the byte array of the file part being uploaded.
         :param body: The request body
         :param filename: Specify the name of the file being uploaded. If no attachment name is provided, the file name is used.
         :param busobid: Specify the Business Object ID.
         :param busobrecid: Specify the Business Object record ID to attach the file to.
         :param offset: The offset is the starting index of the file part being uploaded.  If this is the first part then the offset will be zero.
         :param totalsize: The entire file size in bytes.
         :param attachmentid: Specify the attachment ID of an uploaded file to upload subsequent parts and ensure each part gets appended to the parts that have already been uploaded.
         :param displaytext: Specify the attachment name, which is the display text for the attachment record.
         :return: str"""
        self.validate_path_param(filename)
        self.validate_path_param(busobid)
        self.validate_path_param(busobrecid)
        params = {}
        if attachmentid is not None:
            params["attachmentid"] = attachmentid
        if displaytext is not None:
            params["displaytext"] = displaytext
        return await self.post_body(
            f"/api/V1/uploadbusinessobjectattachment/filename/{filename}/busobid/{busobid}/busobrecid/{busobrecid}/offset/{offset}/totalsize/{totalsize}",
            params=params,
            content=body,
            content_type="application/octet-stream",
        )

    async def UploadBusinessObjectAttachmentByNameAndPublicIdV1(
        self,
        body: str,
        filename: str,
        busobname: str,
        publicid: str,
        offset: int,
        totalsize: int,
        attachmentid: Optional[str] = None,
        displaytext: Optional[str] = None,
    ) -> str:
        """Upload an attachment by Business Object name and public ID

        Operation to upload an attachment to a Business Object record using a Business Object name and public ID. The body of the request is the byte array of the file part being uploaded.
         :param body: The request body
         :param filename: Specify the name of the file being uploaded. If no attachment name is provided, the file name is used.
         :param busobname: Specify the Business Object name.
         :param publicid: Specify the Business Object public ID  to attach the file to.
         :param offset: The offset is the starting index of the file part being uploaded.  If this is the first part then the offset will be zero.
         :param totalsize: The entire file size in bytes.
         :param attachmentid: Specify the attachment ID of an uploaded file to upload subsequent parts and ensure each part gets appended to the parts that have already been uploaded.
         :param displaytext: Specify the attachment name, which is the display text for the attachment record.
         :return: str"""
        self.validate_path_param(filename)
        self.validate_path_param(busobname)
        self.validate_path_param(publicid)
        params = {}
        if attachmentid is not None:
            params["attachmentid"] = attachmentid
        if displaytext is not None:
            params["displaytext"] = displaytext
        return await self.post_body(
            f"/api/V1/uploadbusinessobjectattachment/filename/{filename}/busobname/{busobname}/publicid/{publicid}/offset/{offset}/totalsize/{totalsize}",
            params=params,
            content=body,
            content_type="application/octet-stream",
        )

    async def UploadBusinessObjectAttachmentByNameAndRecIdV1(
        self,
        body: str,
        filename: str,
        busobname: str,
        busobrecid: cherwell_pydantic_api.types.BusObRecID,
        offset: int,
        totalsize: int,
        attachmentid: Optional[str] = None,
        displaytext: Optional[str] = None,
    ) -> str:
        """Upload an attachment by Business Object name and record ID

        Operation to upload an attachment to a Business Object record using a Business Object name and record ID. The body of the request is the byte array of the file part being uploaded.
         :param body: The request body
         :param filename: Specify the name of the file being uploaded. If no attachment name is provided, the file name is used.
         :param busobname: Specify the Business Object name.
         :param busobrecid: Specify the Business Object record ID to attach the file to.
         :param offset: The offset is the starting index of the file part being uploaded.  If this is the first part then the offset will be zero.
         :param totalsize: The entire file size in bytes.
         :param attachmentid: Specify the attachment ID of an uploaded file to upload subsequent parts and ensure each part gets appended to the parts that have already been uploaded.
         :param displaytext: Specify the attachment name, which is the display text for the attachment record.
         :return: str"""
        self.validate_path_param(filename)
        self.validate_path_param(busobname)
        self.validate_path_param(busobrecid)
        params = {}
        if attachmentid is not None:
            params["attachmentid"] = attachmentid
        if displaytext is not None:
            params["displaytext"] = displaytext
        return await self.post_body(
            f"/api/V1/uploadbusinessobjectattachment/filename/{filename}/busobname/{busobname}/busobrecid/{busobrecid}/offset/{offset}/totalsize/{totalsize}",
            params=params,
            content=body,
            content_type="application/octet-stream",
        )



# generated by datamodel-codegen:
#   filename:  swagger.json
#   timestamp: 2023-03-19T19:24:29+00:00

from __future__ import annotations

from enum import Enum
from typing import Optional

import cherwell_pydantic_api.types as ct
from cherwell_pydantic_api.generated_api_utils import ApiBaseModel


class AddItemToQueueRequest(ApiBaseModel):
    busObId: Optional[ct.BusObID] = None
    busObRecId: Optional[ct.BusObRecID] = None
    queueStandInKey: Optional[str] = None


import cherwell_pydantic_api.types as ct
from cherwell_pydantic_api.generated_api_utils import HttpStatusCode


class AddItemToQueueResponse(ApiBaseModel):
    historyRecId: Optional[str] = None
    historyText: Optional[str] = None
    historyTypeId: Optional[str] = None
    errorCode: Optional[str] = None
    errorMessage: Optional[str] = None
    hasError: Optional[bool] = None
    httpStatusCode: Optional[HttpStatusCode] = None


import cherwell_pydantic_api.types as ct


class CheckInQueueItemRequest(ApiBaseModel):
    busObId: Optional[ct.BusObID] = None
    busObRecId: Optional[ct.BusObRecID] = None
    historyNotes: Optional[str] = None
    queueStandInKey: Optional[str] = None


import cherwell_pydantic_api.types as ct


class CheckInQueueItemResponse(ApiBaseModel):
    historyRecId: Optional[str] = None
    historyText: Optional[str] = None
    historyTypeId: Optional[str] = None
    errorCode: Optional[str] = None
    errorMessage: Optional[str] = None
    hasError: Optional[bool] = None
    httpStatusCode: Optional[HttpStatusCode] = None


import cherwell_pydantic_api.types as ct


class CheckOutQueueItemRequest(ApiBaseModel):
    busObId: Optional[ct.BusObID] = None
    busObRecId: Optional[ct.BusObRecID] = None
    historyNotes: Optional[str] = None
    queueStandInKey: Optional[str] = None


import cherwell_pydantic_api.types as ct


class CheckOutQueueItemResponse(ApiBaseModel):
    historyRecId: Optional[str] = None
    historyText: Optional[str] = None
    historyTypeId: Optional[str] = None
    errorCode: Optional[str] = None
    errorMessage: Optional[str] = None
    hasError: Optional[bool] = None
    httpStatusCode: Optional[HttpStatusCode] = None


import cherwell_pydantic_api.types as ct


class RemoveItemFromQueueRequest(ApiBaseModel):
    busObId: Optional[ct.BusObID] = None
    busObRecId: Optional[ct.BusObRecID] = None
    historyNotes: Optional[str] = None
    queueStandInKey: Optional[str] = None


import cherwell_pydantic_api.types as ct


class RemoveItemFromQueueResponse(ApiBaseModel):
    historyRecId: Optional[str] = None
    historyText: Optional[str] = None
    historyTypeId: Optional[str] = None
    errorCode: Optional[str] = None
    errorMessage: Optional[str] = None
    hasError: Optional[bool] = None
    httpStatusCode: Optional[HttpStatusCode] = None



from typing import TYPE_CHECKING, Optional



# This interface doesn't actually do anything, but is used for type checking.

if TYPE_CHECKING:
    from cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.BusinessObject import ReadResponse, SchemaResponse
    from cherwell_pydantic_api.types import BusObID, BusObRecID
    from cherwell_pydantic_api.settings import InstanceSettingsBase
    
    class ApiRequesterInterface:
        async def get_busobid(self, busobname: str) -> Optional[BusObID]:
            ...

        async def get_bo_schema(self, *, busobid: Optional[BusObID] = None, busobname: Optional[str] = None) -> Optional[SchemaResponse]:
            ...

        async def get_bo(self, busobid: BusObID, *, busobrecid: Optional[BusObRecID] = None, publicid: Optional[str] = None) -> ReadResponse:
            ...
        @property
        def settings(self) -> InstanceSettingsBase:
            ...

else:
    class ApiRequesterInterface:
        pass


from typing import Any, Dict, List, Literal, Optional, Union, AsyncIterable, Iterable
from pydantic import parse_obj_as
import cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts
import cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.Lifecycle
from cherwell_pydantic_api.generated_api_utils import GeneratedInterfaceBase


class LifecycleInterface(GeneratedInterfaceBase):
    async def GetStages(
        self,
        businessObjectDefinitionId: str,
    ) -> cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.Lifecycle.GetStagesResponse:
        """Get lifecycle stages

        Gets all of the stages on the lifecycle for a Business Object
         :param businessObjectDefinitionId:
         :return: cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.Lifecycle.GetStagesResponse
        """
        self.validate_path_param(businessObjectDefinitionId)
        response = await self.get(
            f"/api/V1/{businessObjectDefinitionId}/lifecycle/stages"
        )
        return parse_obj_as(
            cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.Lifecycle.GetStagesResponse,
            response,
        )

    async def GetStatuses(
        self,
        businessObjectDefinitionId: str,
    ) -> cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.Lifecycle.GetStatusesResponse:
        """Get lifecycle statuses

        Gets all of the statuses on the lifecycle for a Business Object
         :param businessObjectDefinitionId:
         :return: cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.Lifecycle.GetStatusesResponse
        """
        self.validate_path_param(businessObjectDefinitionId)
        response = await self.get(
            f"/api/V1/{businessObjectDefinitionId}/lifecycle/statuses"
        )
        return parse_obj_as(
            cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.Lifecycle.GetStatusesResponse,
            response,
        )

    async def GetTransitions(
        self,
        businessObjectDefinitionId: str,
    ) -> cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.Lifecycle.GetTransitionsResponse:
        """Get lifecycle transitions

        Gets all of the transitions on the lifecycle for a Business Object
         :param businessObjectDefinitionId:
         :return: cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.Lifecycle.GetTransitionsResponse
        """
        self.validate_path_param(businessObjectDefinitionId)
        response = await self.get(
            f"/api/V1/{businessObjectDefinitionId}/lifecycle/transitions"
        )
        return parse_obj_as(
            cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.Lifecycle.GetTransitionsResponse,
            response,
        )

    async def GetRecordStatus(
        self,
        businessObjectDefinitionId: str,
        recordId: str,
    ) -> cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.Lifecycle.GetRecordStatusResponse:
        """Get current lifecycle status for record

        Gets the current lifecycle status of a business object record
         :param businessObjectDefinitionId:
         :param recordId:
         :return: cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.Lifecycle.GetRecordStatusResponse
        """
        self.validate_path_param(businessObjectDefinitionId)
        self.validate_path_param(recordId)
        response = await self.get(
            f"/api/V1/{businessObjectDefinitionId}/records/{recordId}/status"
        )
        return parse_obj_as(
            cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.Lifecycle.GetRecordStatusResponse,
            response,
        )

    async def GetRecordStage(
        self,
        businessObjectDefinitionId: str,
        recordId: str,
    ) -> cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.Lifecycle.GetRecordStatusResponse:
        """Get current lifecycle stage for record

        Gets the current lifecycle stage of a business object record
         :param businessObjectDefinitionId:
         :param recordId:
         :return: cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.Lifecycle.GetRecordStatusResponse
        """
        self.validate_path_param(businessObjectDefinitionId)
        self.validate_path_param(recordId)
        response = await self.get(
            f"/api/V1/{businessObjectDefinitionId}/records/{recordId}/stage"
        )
        return parse_obj_as(
            cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.Lifecycle.GetRecordStatusResponse,
            response,
        )

    async def GetTransitionOptions(
        self,
        businessObjectDefinitionId: str,
        recordId: str,
    ) -> cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.Lifecycle.GetTransitionOptionsResponse:
        """Get lifecycle transition options for record

        Gets the lifecycle transition options currently available to a business object record
         :param businessObjectDefinitionId:
         :param recordId:
         :return: cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.Lifecycle.GetTransitionOptionsResponse
        """
        self.validate_path_param(businessObjectDefinitionId)
        self.validate_path_param(recordId)
        response = await self.get(
            f"/api/V1/{businessObjectDefinitionId}/records/{recordId}/transitionOptions"
        )
        return parse_obj_as(
            cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.Lifecycle.GetTransitionOptionsResponse,
            response,
        )

    async def TransitionRecord(
        self,
        businessObjectDefinitionId: str,
        recordId: str,
        transitionRecordRequest: cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.Lifecycle.TransitionRecordRequest,
    ) -> cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.ResponseBase:
        """Transition a business object record

        Transitions a business object record in to the specified lifecycle status
         :param businessObjectDefinitionId:
         :param recordId:
         :param transitionRecordRequest: The request body
         :return: cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.ResponseBase
        """
        self.validate_path_param(businessObjectDefinitionId)
        self.validate_path_param(recordId)
        response = await self.post_body(
            f"/api/V1/{businessObjectDefinitionId}/records/{recordId}/transitions",
            content=transitionRecordRequest.json(exclude_unset=True, by_alias=True),
        )
        return parse_obj_as(
            cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.ResponseBase,
            response,
        )



# generated by datamodel-codegen:
#   filename:  swagger.json
#   timestamp: 2023-03-19T19:24:29+00:00

from __future__ import annotations

from enum import Enum
from typing import Optional

import cherwell_pydantic_api.types as ct
from cherwell_pydantic_api.generated_api_utils import ApiBaseModel, HttpStatusCode


class ResponseBase(ApiBaseModel):
    errorCode: Optional[str] = None
    errorMessage: Optional[str] = None
    hasError: Optional[bool] = None
    httpStatusCode: Optional[HttpStatusCode] = None
